"""
Functions and methods for creating ancillary quality control variables
and filters (masks) which can be used with various corrections
routines in ACT.

"""

import dask
import numpy as np
import xarray as xr

from act.qc import comparison_tests, qctests, bsrn_tests, qc_summary
from act.utils.data_utils import get_missing_value


@xr.register_dataset_accessor('qcfilter')
class QCFilter(qctests.QCTests, comparison_tests.QCTests, bsrn_tests.QCTests, qc_summary.QCSummary):
    """
    A class for building quality control variables containing arrays for
    filtering data based on a set of test condition typically based on the
    values in the data fields. These filters can be used in various
    algorithms and calculations within ACT.

    """

    def __init__(self, ds):
        """initialize"""
        self._ds = ds

    def check_for_ancillary_qc(self, var_name, add_if_missing=True, cleanup=False, flag_type=False):
        """
        Method to check if a quality control variable exist in the dataset
        and return the quality control varible name.
        Will call create_qc_variable() to make variable if does not exist
        and update_ancillary_variable() to ensure linkage between data and
        quality control variable. Can also be used just to get the
        corresponding quality control variable name with adding if
        it is missing.

        Parameters
        ----------
        var_name : str
            Data variable name.
        add_if_missing : boolean
            Add quality control variable if missing from teh dataset. Will raise
            and exception if the var_name does not exist in Dataset. Set to False
            to not raise exception.
        cleanup : boolean
            Option to run qc.clean.cleanup() method on the dataset
            to ensure the dataset was updated from ARM QC to the
            correct standardized QC.
        flag_type : boolean
            Indicating the QC variable uses flag_values instead of
            flag_masks.

        Returns
        -------
        qc_var_name : str or None
            Name of existing or new quality control variable. Returns
            None if no existing quality control variable is found and
            add_if_missing is set to False.

        Examples
        --------
            .. code-block:: python

                from act.tests import EXAMPLE_METE40
                from act.io.arm import read_arm_netcdf
                ds = read_arm_netcdf(EXAMPLE_METE40, cleanup_qc=True)
                qc_var_name = ds.qcfilter.check_for_ancillary_qc('atmos_pressure')
                print(f'qc_var_name: {qc_var_name}')
                qc_var_name = ds.qcfilter.check_for_ancillary_qc('the_greatest_variable_ever',
                    add_if_missing=False)
                print(f'qc_var_name: {qc_var_name}')

        """
        qc_var_name = None
        try:
            ancillary_variables = self._ds[var_name].attrs['ancillary_variables']
            var_dims = self._ds[var_name].dims
            if isinstance(ancillary_variables, str):
                ancillary_variables = ancillary_variables.split()
            for var in ancillary_variables:
                for attr, value in self._ds[var].attrs.items():
                    if attr == 'standard_name' and 'quality_flag' in value:
                        if var_dims == self._ds[var].dims:
                            qc_var_name = var

            if add_if_missing and qc_var_name is None:
                qc_var_name = self._ds.qcfilter.create_qc_variable(var_name, flag_type=flag_type)

        except KeyError:
            # Since no ancillary_variables exist look for ARM style of QC
            # variable name. If it exists use it else create new
            # QC varaible.
            if add_if_missing:
                try:
                    self._ds['qc_' + var_name]
                    qc_var_name = 'qc_' + var_name
                except KeyError:
                    qc_var_name = self._ds.qcfilter.create_qc_variable(
                        var_name, flag_type=flag_type
                    )

        # Make sure data varaible has a variable attribute linking
        # data variable to QC variable.
        if add_if_missing:
            self._ds.qcfilter.update_ancillary_variable(var_name, qc_var_name)

        # Clean up quality control variables to the requried standard in the
        # xarray dataset.
        if cleanup:
            self._ds.clean.cleanup(handle_missing_value=True, link_qc_variables=False)

        return qc_var_name

    def create_qc_variable(
        self, var_name, flag_type=False, flag_values_set_value=0, qc_var_name=None
    ):
        """
        Method to create a quality control variable in the dataset.
        Will try not to destroy the qc variable by appending numbers
        to the variable name if needed.

        Parameters
        ----------
        var_name : str
            Data variable name.
        flag_type : boolean
            If an integer flag type should be created instead of
            bitpacked mask type. Will create flag_values instead of
            flag_masks.
        flag_values_set_value : int
            Initial flag value to use when initializing array.
        qc_var_name : str
            Optional new quality control variable name. If not set
            will create one using \\"qc\\_\\" prepended to the data
            variable name. If the name given or created is taken
            will append a number that does not have a conflict.

        Returns
        -------
        qc_var_name : str
            Name of new quality control variable created.

        Examples
        --------
            .. code-block:: python

                from act.tests import EXAMPLE_AOSMET
                from act.io.arm import read_arm_netcdf
                ds = read_arm_netcdf(EXAMPLE_AOSMET)
                qc_var_name = ds.qcfilter.create_qc_variable('temperature_ambient')
                print(qc_var_name)
                print(ds[qc_var_name])

        """

        # Make QC variable long name. The variable long_name attribute
        # may not exist so catch that error and set to default.
        try:
            qc_variable_long_name = (
                'Quality check results on field: ' + self._ds[var_name].attrs['long_name']
            )
        except KeyError:
            qc_variable_long_name = 'Quality check results for ' + var_name

        # Make a new quality control variable name. Check if exists in the
        # dataset. If so loop through creation of new name until one is
        # found that will not replace existing variable.
        if qc_var_name is None:
            qc_var_name = 'qc_' + var_name

        variable_names = list(self._ds.data_vars)
        if qc_var_name in variable_names:
            for ii in range(1, 100):
                temp_qc_var_name = '_'.join([qc_var_name, str(ii)])
                if temp_qc_var_name not in variable_names:
                    qc_var_name = temp_qc_var_name
                    break

        # Create the QC variable filled with 0 values matching the
        # shape of data variable.
        try:
            qc_data = dask.array.from_array(
                np.zeros_like(self._ds[var_name].values, dtype=np.int32),
                chunks=self._ds[var_name].data.chunksize,
            )
        except AttributeError:
            qc_data = np.zeros_like(self._ds[var_name].values, dtype=np.int32)

        # Updating to use coords instead of dim, which caused a loss of
        # attribuets as noted in Issue 347
        self._ds[qc_var_name] = xr.DataArray(
            data=qc_data,
            coords=self._ds[var_name].coords,
            attrs={'long_name': qc_variable_long_name, 'units': '1'},
        )

        # Update if using flag_values and don't want 0 to be default value.
        if flag_type and flag_values_set_value != 0:
            self._ds[qc_var_name].values = self._ds[qc_var_name].values + int(flag_values_set_value)

        # Add requried variable attributes.
        if flag_type:
            self._ds[qc_var_name].attrs['flag_values'] = []
        else:
            self._ds[qc_var_name].attrs['flag_masks'] = []
        self._ds[qc_var_name].attrs['flag_meanings'] = []
        self._ds[qc_var_name].attrs['flag_assessments'] = []
        self._ds[qc_var_name].attrs['standard_name'] = 'quality_flag'

        self.update_ancillary_variable(var_name, qc_var_name=qc_var_name)

        return qc_var_name

    def update_ancillary_variable(self, var_name, qc_var_name=None):
        """
        Method to check if ancillary_variables variable attribute
        is set with quality control variable name.

        Parameters
        ----------
        var_name : str
            Data variable name.
        qc_var_name : str
            quality control variable name. If not given will attempt
            to get the name from data variable ancillary_variables
            attribute.

        Examples
        --------
            .. code-block:: python

                from act.tests import EXAMPLE_AOSMET
                from act.io.arm import read_arm_netcdf
                ds = read_arm_netcdf(EXAMPLE_AOSMET)
                var_name = 'temperature_ambient'
                qc_var_name = ds.qcfilter.create_qc_variable(var_name)
                del ds[var_name].attrs['ancillary_variables']
                ds.qcfilter.update_ancillary_variable(var_name, qc_var_name)
                print(ds[var_name].attrs['ancillary_variables'])

        """
        if qc_var_name is None:
            qc_var_name = self._ds.qcfilter.check_for_ancillary_qc(var_name, add_if_missing=False)

        if qc_var_name is None:
            return

        try:
            ancillary_variables = self._ds[var_name].attrs['ancillary_variables']
            if qc_var_name not in ancillary_variables:
                ancillary_variables = ' '.join([ancillary_variables, qc_var_name])
        except KeyError:
            ancillary_variables = qc_var_name

        self._ds[var_name].attrs['ancillary_variables'] = ancillary_variables

    def add_test(
        self,
        var_name,
        index=None,
        test_number=None,
        test_meaning=None,
        test_assessment='Bad',
        flag_value=False,
        recycle=False,
    ):
        """
        Method to add a new test/filter to a quality control variable.

        Parameters
        ----------
        var_name : str
            data variable name
        index : int, bool, list of int or bool, numpy array, tuple of numpy arrays, None
            Indexes into quality control array to set the test bit.
            If not set or set to None will not set the test on any
            element of the quality control variable but will still
            add the test to the flag_masks, flag_meanings and
            flag_assessments attributes.
        test_number : int
            Test number to use. If keyword is not set will use first
            available test bit/test number.
        test_meaning : str
            String describing the test. Will be added to flag_meanings
            variable attribute.
        test_assessment : str
            String describing the test assessment. If not set will use
            "Bad" as the string to append to flag_assessments. Will
            update to be capitalized.
        flag_value : boolean
            Switch to use flag_values integer quality control.
        recyle : boolean
            Option to use number less than next highest test if available. For example
            tests 1, 2, 4, 5 are set. Set to true the next test chosen will be 3, else
            will be 6.

        Returns
        -------
        test_dict : dict
            A dictionary containing information added to the QC
            variable.

        Examples
        --------
            .. code-block:: python

                result = ds.qcfilter.add_test(var_name, test_meaning='Birds!')

        """
        test_dict = {}

        if test_meaning is None:
            raise ValueError(
                'You need to provide a value for test_meaning '
                'keyword when calling the add_test method'
            )

        # This ensures the indexing will work even if given float values.
        # Preserves tuples from np.where() or boolean arrays for standard
        # python indexing.
        if index is not None and not isinstance(index, (np.ndarray, tuple)):
            index = np.array(index)
            if index.dtype.kind not in np.typecodes['AllInteger']:
                index = index.astype(int)

        # Ensure assessment is capitalized to be consistent
        test_assessment = test_assessment.capitalize()

        qc_var_name = self._ds.qcfilter.check_for_ancillary_qc(var_name, flag_type=flag_value)

        if test_number is None:
            test_number = self._ds.qcfilter.available_bit(qc_var_name, recycle=recycle)

        self._ds.qcfilter.set_test(var_name, index, test_number, flag_value)

        if flag_value:
            try:
                self._ds[qc_var_name].attrs['flag_values'].append(test_number)
            except KeyError:
                self._ds[qc_var_name].attrs['flag_values'] = [test_number]
        else:
            # Determine if flag_masks test number is too large for current data type.
            # If so up convert data type.
            flag_masks = np.array(self._ds[qc_var_name].attrs['flag_masks'])
            mask_dtype = flag_masks.dtype
            if not np.issubdtype(mask_dtype, np.integer):
                mask_dtype = np.uint32

            if np.iinfo(mask_dtype).max - set_bit(0, test_number) <= -1:
                if mask_dtype == np.int8 or mask_dtype == np.uint8:
                    mask_dtype = np.uint16
                elif mask_dtype == np.int16 or mask_dtype == np.uint16:
                    mask_dtype = np.uint32
                elif mask_dtype == np.int32 or mask_dtype == np.uint32:
                    mask_dtype = np.uint64

            flag_masks = flag_masks.astype(mask_dtype)
            flag_masks = np.append(flag_masks, np.array(set_bit(0, test_number), dtype=mask_dtype))
            self._ds[qc_var_name].attrs['flag_masks'] = list(flag_masks)

        try:
            self._ds[qc_var_name].attrs['flag_meanings'].append(test_meaning)
        except KeyError:
            self._ds[qc_var_name].attrs['flag_meanings'] = [test_meaning]

        try:
            self._ds[qc_var_name].attrs['flag_assessments'].append(test_assessment)
        except KeyError:
            self._ds[qc_var_name].attrs['flag_assessments'] = [test_assessment]

        test_dict['test_number'] = test_number
        test_dict['test_meaning'] = test_meaning
        test_dict['test_assessment'] = test_assessment
        test_dict['qc_variable_name'] = qc_var_name
        test_dict['variable_name'] = var_name

        return test_dict

    def remove_test(
        self,
        var_name=None,
        test_number=None,
        qc_var_name=None,
        flag_value=False,
        flag_values_reset_value=0,
    ):
        """
        Method to remove a test/filter from a quality control variable. Must set
        var_name or qc_var_name.

        Parameters
        ----------
        var_name : str or None
            Data variable name.
        test_number : int
            Test number to remove.
        qc_var_name : str or None
            Quality control variable name. Ignored if var_name is set.
        flag_value : boolean
            Switch to use flag_values integer quality control.
        flag_values_reset_value : int
            Value to use when resetting a flag_values value to not be set.

        Examples
        --------
            .. code-block:: python

                ds.qcfilter.remove_test(var_name, test_number=3)

        """
        if test_number is None:
            raise ValueError(
                'You need to provide a value for test_number '
                'keyword when calling the remove_test() method'
            )

        if var_name is None and qc_var_name is None:
            raise ValueError(
                'You need to provide a value for var_name or qc_var_name '
                'keyword when calling the remove_test() method'
            )

        if var_name is not None:
            qc_var_name = self._ds.qcfilter.check_for_ancillary_qc(var_name)

        # Determine which index is using the test number
        index = None
        if flag_value:
            flag_values = self._ds[qc_var_name].attrs['flag_values']
            for ii, flag_num in enumerate(flag_values):
                if flag_num == test_number:
                    index = ii
                    break
        else:
            flag_masks = self._ds[qc_var_name].attrs['flag_masks']
            for ii, bit_num in enumerate(flag_masks):
                if parse_bit(bit_num)[0] == test_number:
                    index = ii
                    break

        # If can't find the index of test return before doing anything.
        if index is None:
            return

        if flag_value:
            remove_index = self._ds.qcfilter.get_qc_test_mask(
                var_name=var_name,
                qc_var_name=qc_var_name,
                test_number=test_number,
                return_index=True,
                flag_value=True,
            )
            self._ds.qcfilter.unset_test(
                var_name=var_name,
                qc_var_name=qc_var_name,
                index=remove_index,
                test_number=test_number,
                flag_value=flag_value,
                flag_values_reset_value=flag_values_reset_value,
            )
            del flag_values[index]
            self._ds[qc_var_name].attrs['flag_values'] = flag_values

        else:
            remove_index = self._ds.qcfilter.get_qc_test_mask(
                var_name=var_name,
                qc_var_name=qc_var_name,
                test_number=test_number,
                return_index=True,
            )
            self._ds.qcfilter.unset_test(
                var_name=var_name,
                qc_var_name=qc_var_name,
                index=remove_index,
                test_number=test_number,
                flag_value=flag_value,
            )
            if isinstance(flag_masks, list):
                del flag_masks[index]
            else:
                flag_masks = np.delete(flag_masks, index)
            self._ds[qc_var_name].attrs['flag_masks'] = flag_masks

        flag_meanings = self._ds[qc_var_name].attrs['flag_meanings']
        del flag_meanings[index]
        self._ds[qc_var_name].attrs['flag_meanings'] = flag_meanings

        flag_assessments = self._ds[qc_var_name].attrs['flag_assessments']
        del flag_assessments[index]
        self._ds[qc_var_name].attrs['flag_assessments'] = flag_assessments

    def set_test(self, var_name, index=None, test_number=None, flag_value=False):
        """
        Method to set a test/filter in a quality control variable.

        Parameters
        ----------
        var_name : str
            Data variable name.
        index : int or list or numpy array
            Index to set test in quality control array. If want to
            unset all values will need to pass in index of all values.
        test_number : int
            Test number to set.
        flag_value : boolean
            Switch to use flag_values integer quality control.

        Examples
        --------
            .. code-block:: python

                index = [0, 1, 2, 30]
                ds.qcfilter.set_test(var_name, index=index, test_number=2)

        """

        qc_var_name = self._ds.qcfilter.check_for_ancillary_qc(var_name)

        qc_variable = np.array(self._ds[qc_var_name].values)

        # Ensure the qc_variable data type is integer. This ensures bitwise comparison
        # will not cause an error.
        if qc_variable.dtype.kind not in np.typecodes['AllInteger']:
            qc_variable = qc_variable.astype(int)

        # Determine if test number is too large for current data type. If so
        # up convert data type.
        dtype = qc_variable.dtype
        if np.iinfo(dtype).max - set_bit(0, test_number) < -1:
            if dtype == np.int8:
                dtype = np.int16
            elif dtype == np.int16:
                dtype = np.int32
            elif dtype == np.int32:
                dtype = np.int64

            qc_variable = qc_variable.astype(dtype)

        if index is not None:
            if flag_value:
                if len(qc_variable.shape) == 0:
                    qc_variable = test_number
                else:
                    qc_variable[index] = test_number
            else:
                if bool(np.shape(index)):
                    qc_variable[index] = set_bit(qc_variable[index], test_number)
                elif index == 0:
                    qc_variable = set_bit(qc_variable, test_number)

        self._ds[qc_var_name].values = qc_variable

    def unset_test(
        self,
        var_name=None,
        qc_var_name=None,
        index=None,
        test_number=None,
        flag_value=False,
        flag_values_reset_value=0,
    ):
        """
        Method to unset a test/filter from a quality control variable.

        Parameters
        ----------
        var_name : str or None
            Data variable name.
        qc_var_name : str or None
            Quality control variable name. Ignored if var_name is set.
        index : int or list or numpy array
            Index to unset test in quality control array. If want to
            unset all values will need to pass in index of all values.
        test_number : int
            Test number to remove.
        flag_value : boolean
            Switch to use flag_values integer quality control.
        flag_values_reset_value : int
            Value to use when resetting a flag_values value to not be set.

        Examples
        --------
        .. code-block:: python

            ds.qcfilter.unset_test(var_name, index=range(10, 100), test_number=2)

        """
        if index is None:
            return

        if var_name is None and qc_var_name is None:
            raise ValueError(
                'You need to provide a value for var_name or qc_var_name '
                'keyword when calling the unset_test() method'
            )

        if var_name is not None:
            qc_var_name = self._ds.qcfilter.check_for_ancillary_qc(var_name)

        # Get QC variable
        qc_variable = self._ds[qc_var_name].values

        # Ensure the qc_variable data type is integer. This ensures bitwise comparison
        # will not cause an error.
        if qc_variable.dtype.kind not in np.typecodes['AllInteger']:
            qc_variable = qc_variable.astype(int)

        if flag_value:
            qc_variable[index] = flag_values_reset_value
        else:
            qc_variable[index] = unset_bit(qc_variable[index], test_number)

        self._ds[qc_var_name].values = qc_variable

    def available_bit(self, qc_var_name, recycle=False):
        """
        Method to determine next available bit or flag to use with a QC test.
        Will check for flag_masks first and if not found will check for
        flag_values. This will drive how the next value is chosen.

        Parameters
        ----------
        qc_var_name : str
            Quality control variable name.
        recycle : boolean
            Option to look for a bit (test) not in use starting from 1.
            If a test is not defined will return the lowest number, else
            will just use next highest number.

        Returns
        -------
        test_num : int
            Next available test number.

        Examples
        --------
            .. code-block:: python

                from act.tests import EXAMPLE_METE40
                from act.io.arm import read_arm_netcdf
                ds = read_arm_netcdf(EXAMPLE_METE40, cleanup_qc=True)
                test_number = ds.qcfilter.available_bit('qc_atmos_pressure')
                print(test_number)


        """
        try:
            flag_masks = self._ds[qc_var_name].attrs['flag_masks']
            flag_value = False
        except KeyError:
            try:
                flag_masks = self._ds[qc_var_name].attrs['flag_values']
                flag_value = True
            except KeyError:
                try:
                    self._ds[qc_var_name].attrs['flag_values']
                    flag_masks = self._ds[qc_var_name].attrs['flag_masks']
                    flag_value = False
                except KeyError:
                    raise ValueError(
                        'Problem getting next value from '
                        'available_bit(). flag_values and '
                        'flag_masks not set as expected'
                    )

        if flag_masks == []:
            next_bit = 1
        else:
            if flag_value:
                if recycle:
                    next_bit = min(set(range(1, 100000)) - set(flag_masks))
                else:
                    next_bit = max(flag_masks) + 1
            else:
                if recycle:
                    tests = [parse_bit(mask)[0] for mask in flag_masks]
                    next_bit = min(set(range(1, 63)) - set(tests))
                else:
                    next_bit = parse_bit(max(flag_masks))[0] + 1

        return int(next_bit)

    def get_qc_test_mask(
        self,
        var_name=None,
        test_number=None,
        qc_var_name=None,
        flag_value=False,
        return_index=False,
    ):
        """
        Returns a numpy array of False or True where a particular
        flag or bit is set in a numpy array. Must set var_name or qc_var_name
        when calling.

        Parameters
        ----------
        var_name : str or None
            Data variable name.
        test_number : int
            Test number to return array where test is set.
        qc_var_name : str or None
            Quality control variable name. Ignored if var_name is set.
        flag_value : boolean
            Switch to use flag_values integer quality control.
        return_index : boolean
            Return a numpy array of index numbers into QC array where the
            test is set instead of False or True mask.

        Returns
        -------
        test_mask : numpy bool array or numpy integer array
            A numpy boolean array with False or True where the test number or
            bit was set, or numpy integer array of indexes where test is True.

        Examples
        --------
            .. code-block:: python

                from act.io.arm import read_arm_netcdf
                from act.tests import EXAMPLE_IRT25m20s

                ds = read_arm_netcdf(EXAMPLE_IRT25m20s)
                var_name = "inst_up_long_dome_resist"
                result = ds.qcfilter.add_test(
                    var_name, index=[0, 1, 2], test_meaning="Birds!"
                )
                qc_var_name = result["qc_variable_name"]
                mask = ds.qcfilter.get_qc_test_mask(
                    var_name, result["test_number"], return_index=True
                )
                print(mask)
                array([0, 1, 2])

                mask = ds.qcfilter.get_qc_test_mask(var_name, result["test_number"])
                print(mask)
                array([True, True, True, ..., False, False, False])

                data = ds[var_name].values
                print(data[mask])
                array([7.84, 7.8777, 7.8965], dtype=float32)

                import numpy as np

                data[mask] = np.nan
                print(data)
                array([nan, nan, nan, ..., 7.6705, 7.6892, 7.6892], dtype=float32)

        """
        if var_name is None and qc_var_name is None:
            raise ValueError(
                'You need to provide a value for var_name or qc_var_name '
                'keyword when calling the get_qc_test_mask() method'
            )

        if test_number is None:
            raise ValueError(
                'You need to provide a value for test_number '
                'keyword when calling the get_qc_test_mask() method'
            )

        if var_name is not None:
            qc_var_name = self._ds.qcfilter.check_for_ancillary_qc(var_name)

        qc_variable = self._ds[qc_var_name].values
        # Ensure the qc_variable data type is integer. This ensures bitwise comparison
        # will not cause an error.
        if qc_variable.dtype.kind not in np.typecodes['AllInteger']:
            qc_variable = qc_variable.astype(int)

        if flag_value:
            tripped = qc_variable == test_number
        else:
            check_bit = set_bit(0, test_number) & qc_variable
            tripped = check_bit > 0

        test_mask = np.full(qc_variable.shape, False, dtype='bool')
        # Make sure test_mask is an array. If qc_variable is scalar will
        # be retuned from np.zeros as scalar.
        test_mask = np.atleast_1d(test_mask)
        test_mask[tripped] = True
        test_mask = np.ma.make_mask(test_mask, shrink=False)

        if return_index:
            test_mask = np.where(test_mask)[0]

        return test_mask

    def get_masked_data(
        self,
        var_name,
        rm_assessments=None,
        rm_tests=None,
        return_nan_array=False,
        ma_fill_value=None,
        return_inverse=False,
        return_mask_only=False,
    ):
        """
        Returns a numpy masked array containing data and mask or
        a numpy float array with masked values set to NaN.

        Parameters
        ----------
        var_name : str
            Data variable name.
        rm_assessments : str or list of str
            Assessment name to exclude from returned data.
        rm_tests : int or list of int
            Test numbers to exclude from returned data. This is the test
            number (or bit position number) not the mask number.
        return_nan_array : boolean
            Return a numpy array with filtered ( or masked) values
            set to numpy NaN value. If the data is type int will upconvert
            to numpy float to allow setting NaN value.
        ma_fill_value : int or float (or str?)
            The numpy masked array fill_value used in creation of the the
            masked array. If the datatype needs to be upconverted to allow
            the fill value to be used, data will be upconverted.
        return_inverse : boolean
            Invert the masked array mask or return data array where mask is set
            to False instead of True set to NaN. Useful for overplotting
            where failing.
        return_mask_only : boolean
            Return the boolean mask only as a numpy array.

        Returns
        -------
        variable : numpy masked array or numpy float array
            Default is to return a numpy masked array with the mask set to
            True where the test with requested assessment or test number
            was found set.
            If return_nan_array is True will return numpy array upconverted
            to float with locations where the test with requested assessment
            or test number was found set converted to NaN.

        Examples
        --------
            .. code-block:: python

                from act.io.arm import read_arm_netcdf
                from act.tests import EXAMPLE_IRT25m20s

                ds = read_arm_netcdf(EXAMPLE_IRT25m20s)
                var_name = "inst_up_long_dome_resist"
                result = ds.qcfilter.add_test(
                    var_name, index=[0, 1, 2], test_meaning="Birds!"
                )
                data = ds.qcfilter.get_masked_data(
                    var_name, rm_assessments=["Bad", "Indeterminate"]
                )
                print(data)
                masked_array(
                    data=[..., 7.670499801635742, 7.689199924468994, 7.689199924468994],
                    mask=[..., False, False, False],
                    fill_value=1e20,
                    dtype=float32,
                )

        """
        qc_var_name = self._ds.qcfilter.check_for_ancillary_qc(var_name, add_if_missing=False)

        flag_value = False
        flag_values = None
        flag_masks = None
        flag_assessments = None
        try:
            flag_assessments = self._ds[qc_var_name].attrs['flag_assessments']
            flag_masks = self._ds[qc_var_name].attrs['flag_masks']
        except KeyError:
            pass

        try:
            flag_values = self._ds[qc_var_name].attrs['flag_values']
            flag_value = True
        except KeyError:
            pass

        test_numbers = []
        if rm_tests is not None:
            if isinstance(rm_tests, (int, float, str)):
                rm_tests = [int(rm_tests)]
            test_numbers.extend(rm_tests)

        if rm_assessments is not None:
            if isinstance(rm_assessments, str):
                rm_assessments = [rm_assessments]

            if flag_masks is not None:
                test_nums = [parse_bit(mask)[0] for mask in flag_masks]

            if flag_values is not None:
                test_nums = flag_values

            rm_assessments = [x.lower() for x in rm_assessments]
            if flag_assessments is not None:
                for ii, assessment in enumerate(flag_assessments):
                    if assessment.lower() in rm_assessments:
                        test_numbers.append(test_nums[ii])

        # Make the list of test numbers to mask unique
        test_numbers = list(set(test_numbers))

        # Create mask of indexes by looking where each test is set
        variable = self._ds[var_name].values
        nan_dtype = np.float32
        if variable.dtype in (np.float64, np.int64):
            nan_dtype = np.float64

        try:
            # Get shape of mask from QC variable since there is a chance it will
            # be a different shape than data variable.
            mask = np.zeros(self._ds[qc_var_name].shape, dtype=bool)
        except KeyError:
            # If there is no QC variable make mask from shape of data variable.
            mask = np.zeros(self._ds[var_name].shape, dtype=bool)

        for test in test_numbers:
            qc_test_mask = self._ds.qcfilter.get_qc_test_mask(var_name, test, flag_value=flag_value)
            mask = mask | qc_test_mask

        # If requested only return the mask.
        if return_mask_only:
            return mask

        # Convert data numpy array into masked array
        try:
            variable = np.ma.array(variable, mask=mask, fill_value=ma_fill_value)
        except TypeError:
            variable = np.ma.array(
                variable,
                mask=mask,
                fill_value=ma_fill_value,
                dtype=np.array(ma_fill_value).dtype,
            )

        # If requested switch array from where data is not failing tests
        # to where data is failing tests. This can be used when over plotting
        # where the data if failing the tests.
        if return_inverse:
            mask = variable.mask
            mask = np.invert(mask)
            variable.mask = mask

        # If asked to return numpy array with values set to NaN
        if return_nan_array:
            variable = variable.astype(nan_dtype)
            variable = variable.filled(fill_value=np.nan)

        return variable

    def datafilter(
        self,
        variables=None,
        rm_assessments=None,
        rm_tests=None,
        verbose=False,
        del_qc_var=False,
        no_NaN=False,
    ):
        """
        Method to apply quality control variables to data variables by
        changing the data values in the dataset using quality control variables.
        The data is updated with failing data set to
        NaN. This can be used to update the data variable in the xarray
        dataset for use with xarray methods to perform analysis on the data
        since those methods don't read the quality control variables.

        Parameters
        ----------
        variables : None or str or list of str
            Data variable names to process. If set to None will update all
            data variables.
        rm_assessments : str or list of str
            Assessment names listed under quality control varible flag_assessments
            to exclude from returned data. Examples include
            ['Bad', 'Incorrect', 'Indeterminate', 'Suspect']
        rm_tests : int or list of int
            Test numbers listed under quality control variable to exclude from
            returned data. This is the test
            number (or bit position number) not the mask number.
        verbose : boolean
            Print processing information.
        del_qc_var : boolean
            Option to delete quality control variable after processing. Since
            the data values can not be determined after they are set to NaN
            and xarray method processing would also process the quality control
            variables, the default is to remove the quality control data
            variables.  Defaults to False.
        no_NaN : boolean
            Should the returned Xarray Dataset use NaN as the missing value indicator.
            If Xarray did not convert the _FillValue or missing_value to NaN upon reading
            the same missing value indicator should be used. If set to true will try to determine
            the current missing_value or _FillValue set in the file and use that value. If neither
            are set as a variable attribute it will use the default value (most likely -9999).

        Examples
        --------
            .. code-block:: python

                from act.io.arm import read_arm_netcdf
                from act.tests import EXAMPLE_MET1

                ds = read_arm_netcdf(EXAMPLE_MET1)
                ds.clean.cleanup()

                var_name = "atmos_pressure"

                ds_1 = ds.nanmean()

                ds.qcfilter.add_less_test(var_name, 99, test_assessment="Bad")
                ds.qcfilter.datafilter(rm_assessments="Bad")
                ds_2 = ds.nanmean()

                print("All_data =", ds_1[var_name].values)
                All_data = 98.86098
                print("Bad_Removed =", ds_2[var_name].values)
                Bad_Removed = 99.15148

        """

        if rm_assessments is None and rm_tests is None:
            raise ValueError('Need to set rm_assessments or rm_tests option')

        if variables is not None and isinstance(variables, str):
            variables = [variables]

        if variables is None:
            variables = list(self._ds.data_vars)

        for var_name in variables:
            qc_var_name = self.check_for_ancillary_qc(var_name, add_if_missing=False, cleanup=False)
            if qc_var_name is None:
                if verbose:
                    if var_name in ['base_time', 'time_offset']:
                        continue

                    try:
                        if self._ds[var_name].attrs['standard_name'] == 'quality_flag':
                            continue
                    except KeyError:
                        pass

                    print(
                        f'No quality control variable for {var_name} found '
                        f'in call to .qcfilter.datafilter()'
                    )

                continue

            # Need to return data as Numpy array with NaN values. Setting the Dask array
            # to Numpy masked array does not work with other tools.
            data = self.get_masked_data(
                var_name, rm_assessments=rm_assessments, rm_tests=rm_tests, return_nan_array=True
            )

            if no_NaN:
                missing_value = get_missing_value(self._ds, var_name, add_if_missing_in_ds=True)
                index = np.isnan(data)
                if np.any(index):
                    data[index] = missing_value

            # If data was orginally stored as Dask array return values to Dataset as Dask array
            # else set as Numpy array.
            try:
                self._ds[var_name].data = dask.array.from_array(
                    data, chunks=self._ds[var_name].data.chunksize
                )

            except AttributeError:
                self._ds[var_name].values = data

            # Adding information on filtering to history attribute
            flag_masks = None
            flag_assessments = None
            flag_meanings = None
            try:
                flag_assessments = list(self._ds[qc_var_name].attrs['flag_assessments'])
                flag_masks = list(self._ds[qc_var_name].attrs['flag_masks'])
                flag_meanings = list(self._ds[qc_var_name].attrs['flag_meanings'])
            except KeyError:
                pass

            # Add comment to history for each test that's filtered out
            if rm_tests is not None:
                if isinstance(rm_tests, int):
                    rm_tests = [rm_tests]

                for test in rm_tests:
                    try:
                        index = flag_masks.index(set_bit(0, test))
                    except ValueError:
                        continue

                    comment = f'act.qc.datafilter: {flag_meanings[index]}'
                    if 'history' in self._ds[var_name].attrs.keys():
                        self._ds[var_name].attrs['history'] += f'\n {comment}'
                    else:
                        self._ds[var_name].attrs['history'] = comment

            if rm_assessments is not None:
                if isinstance(rm_assessments, str):
                    rm_assessments = [rm_assessments]

                for assessment in rm_assessments:
                    if assessment in flag_assessments:
                        index = [i for i, e in enumerate(flag_assessments) if e == assessment]
                        for ind in index:
                            comment = f'act.qc.datafilter: {flag_meanings[ind]}'
                            if 'history' in self._ds[var_name].attrs.keys():
                                self._ds[var_name].attrs['history'] += f'\n {comment}'
                            else:
                                self._ds[var_name].attrs['history'] = comment

            # If requested delete quality control variable
            if del_qc_var:
                del self._ds[qc_var_name]
                try:
                    if self._ds[var_name].attrs['ancillary_variables'] == qc_var_name:
                        del self._ds[var_name].attrs['ancillary_variables']
                except KeyError:
                    pass

                if verbose:
                    print(f'Deleting {qc_var_name} from dataset')


def set_bit(array, bit_number):
    """
    Function to set a quality control bit given a scalar or
    array of values and a bit number.

    Parameters
    ----------
    array : int list of int or numpy array of int
        The bitpacked array to set the bit number.
    bit_number : int
        The bit (or test) number to set starting at 1.

    Returns
    -------
    array : int, numpy array, tuple, list
        Integer or numpy array with bit set for each element of the array.
        Returned in same type.

    Examples
    --------
    Example use setting bit 2 to an array called data:

        .. code-block:: python

            from act.qc.qcfilter import set_bit
            data = np.array(range(0, 7))
            data = set_bit(data, 2)
            print(data)
            array([2, 3, 2, 3, 6, 7, 6])

    """
    was_list = False
    was_tuple = False
    if isinstance(array, list):
        array = np.array(array)
        was_list = True

    if isinstance(array, tuple):
        array = np.array(array)
        was_tuple = True

    if bit_number > 0:
        array |= 1 << bit_number - 1

    if was_list:
        array = list(array)

    if was_tuple:
        array = tuple(array)

    return array


def unset_bit(array, bit_number):
    """
    Function to remove a quality control bit given a
    scalar or array of values and a bit number.

    Parameters
    ----------
    array : int list of int or numpy array
        Array of integers containing bit packed numbers.
    bit_number : int
        Bit number to remove starting at 1.

    Returns
    -------
    array : int or numpy array
        Returns same data type as array entered with bit removed. Will
        fail gracefully if the bit requested to be removed was not set.

    Examples
    --------
       .. code-block:: python

            from act.qc.qcfilter import set_bit, unset_bit
            data = set_bit([0, 1, 2, 3, 4], 2)
            data = set_bit(data, 3)
            print(data)
            [6, 7, 6, 7, 6]

            data = unset_bit(data, 2)
            print(data)
            [4, 5, 4, 5, 4]

    """
    was_list = False
    was_tuple = False
    if isinstance(array, list):
        array = np.array(array)
        was_list = True

    if isinstance(array, tuple):
        array = np.array(array)
        was_tuple = True

    if bit_number > 0:
        array &= ~(1 << bit_number - 1)

    if was_list:
        array = list(array)

    if was_tuple:
        array = tuple(array)

    return array


def parse_bit(qc_bit):
    """
    Given a single integer value, return bit positions.

    Parameters
    ----------
    qc_bit : int or numpy int
        Bit packed integer number to be parsed.

    Returns
    -------
    bit_number : numpy.int32 array
        Array containing all bit numbers of the bit packed number.
        If no bits set returns empty array.

    Examples
    --------
        .. code-block:: python

            from act.qc.qcfilter import parse_bit
            parse_bit(7)
            array([1, 2, 3], dtype=int32)

    """
    if isinstance(qc_bit, (list, tuple, np.ndarray)):
        if len(qc_bit) > 1:
            raise ValueError('Must be a single value.')
        qc_bit = qc_bit[0]

    if qc_bit < 0:
        raise ValueError('Must be a positive integer.')

    # Convert integer value to single element numpy array of type unsigned integer 64
    value = np.array([qc_bit]).astype(">u8")

    # Convert value to view containing only unsigned integer 8 data type. This
    # is required for the numpy unpackbits function which only works with
    # unsigned integer 8 bit data type.
    value = value.view("u1")

    # Unpack bits using numpy into array of 1 where bit is set and convert into boolean array
    index = np.unpackbits(value).astype(bool)

    # Create range of numbers from 64 to 1 and subset where unpackbits found a bit set.
    bit_number = np.arange(index.size, 0, -1)[index]

    # Flip the array to increasing numbers to match historical method
    bit_number = np.flip(bit_number)

    # bit_number = []
    # qc_bit = int(qc_bit)

    # counter = 0
    # while qc_bit > 0:
    #     temp_value = qc_bit % 2
    #     qc_bit = qc_bit >> 1
    #     counter += 1
    #     if temp_value == 1:
    #         bit_number.append(counter)

    # Convert data type into expected type
    bit_number = np.asarray(bit_number, dtype=np.int32)

    return bit_number



import numpy as np

from act.io.arm import read_arm_netcdf
from act.tests import (
    EXAMPLE_CEIL1,
    EXAMPLE_CO2FLX4M,
    EXAMPLE_MET1,
    EXAMPLE_MET_SAIL,
    EXAMPLE_SIRS_SIRI_QC,
    EXAMPLE_SWATS,
)


def test_global_qc_cleanup():
    ds = read_arm_netcdf(EXAMPLE_MET1)
    ds.load()
    ds.clean.cleanup()

    assert ds['qc_wdir_vec_mean'].attrs['flag_meanings'] == [
        'Value is equal to missing_value.',
        'Value is less than the fail_min.',
        'Value is greater than the fail_max.',
    ]
    assert ds['qc_wdir_vec_mean'].attrs['flag_masks'] == [1, 2, 4]
    assert ds['qc_wdir_vec_mean'].attrs['flag_assessments'] == [
        'Bad',
        'Bad',
        'Bad',
    ]

    assert ds['qc_temp_mean'].attrs['flag_meanings'] == [
        'Value is equal to missing_value.',
        'Value is less than the fail_min.',
        'Value is greater than the fail_max.',
        'Difference between current and previous values exceeds fail_delta.',
    ]
    assert ds['qc_temp_mean'].attrs['flag_masks'] == [1, 2, 4, 8]
    assert ds['qc_temp_mean'].attrs['flag_assessments'] == [
        'Bad',
        'Bad',
        'Bad',
        'Indeterminate',
    ]

    ds.close()
    del ds


def test_clean():
    # Read test data
    ceil_ds = read_arm_netcdf([EXAMPLE_CEIL1])
    # Cleanup QC data
    ceil_ds.clean.cleanup(clean_arm_state_vars=['detection_status'])

    # Check that global attribures are removed
    global_attributes = [
        'qc_bit_comment',
        'qc_bit_1_description',
        'qc_bit_1_assessment',
        'qc_bit_2_description',
        'qc_bit_2_assessment' 'qc_bit_3_description',
        'qc_bit_3_assessment',
    ]

    for glb_att in global_attributes:
        assert glb_att not in ceil_ds.attrs.keys()

    # Check that CF attributes are set including new flag_assessments
    var_name = 'qc_first_cbh'
    for attr_name in ['flag_masks', 'flag_meanings', 'flag_assessments']:
        assert attr_name in ceil_ds[var_name].attrs.keys()
        assert isinstance(ceil_ds[var_name].attrs[attr_name], list)

    # Check that the flag_mask values are set correctly
    assert ceil_ds['qc_first_cbh'].attrs['flag_masks'] == [1, 2, 4]

    # Check that the flag_meanings values are set correctly
    assert ceil_ds['qc_first_cbh'].attrs['flag_meanings'] == [
        'Value is equal to missing_value.',
        'Value is less than the fail_min.',
        'Value is greater than the fail_max.',
    ]

    # Check the value of flag_assessments is as expected
    assert ceil_ds['qc_first_cbh'].attrs['flag_assessments'] == ['Bad', 'Bad', 'Bad']

    # Check that ancillary varibles is being added
    assert 'qc_first_cbh' in ceil_ds['first_cbh'].attrs['ancillary_variables'].split()

    # Check that state field is updated to CF
    assert 'flag_values' in ceil_ds['detection_status'].attrs.keys()
    assert isinstance(ceil_ds['detection_status'].attrs['flag_values'], list)
    assert ceil_ds['detection_status'].attrs['flag_values'] == [0, 1, 2, 3, 4, 5]

    assert 'flag_meanings' in ceil_ds['detection_status'].attrs.keys()
    assert isinstance(ceil_ds['detection_status'].attrs['flag_meanings'], list)
    assert ceil_ds['detection_status'].attrs['flag_meanings'] == [
        'No significant backscatter',
        'One cloud base detected',
        'Two cloud bases detected',
        'Three cloud bases detected',
        'Full obscuration determined but no cloud base detected',
        'Some obscuration detected but determined to be transparent',
    ]

    assert 'flag_0_description' not in ceil_ds['detection_status'].attrs.keys()
    assert 'detection_status' in ceil_ds['first_cbh'].attrs['ancillary_variables'].split()

    ceil_ds.close()


def test_qc_remainder():
    ds = read_arm_netcdf(EXAMPLE_MET1)
    assert ds.clean.get_attr_info(variable='bad_name') is None
    del ds.attrs['qc_bit_comment']
    assert isinstance(ds.clean.get_attr_info(), dict)
    ds.attrs['qc_flag_comment'] = 'testing'
    ds.close()

    ds = read_arm_netcdf(EXAMPLE_MET1)
    ds.clean.cleanup(normalize_assessment=True)
    ds['qc_atmos_pressure'].attrs['units'] = 'testing'
    del ds['qc_temp_mean'].attrs['units']
    del ds['qc_temp_mean'].attrs['flag_masks']
    ds.clean.handle_missing_values()
    ds.close()

    ds = read_arm_netcdf(EXAMPLE_MET1)
    ds.attrs['qc_bit_1_comment'] = 'tesing'
    data = ds['qc_atmos_pressure'].values.astype(np.int64)
    data[0] = 2**32
    ds['qc_atmos_pressure'].values = data
    ds.clean.get_attr_info(variable='qc_atmos_pressure')
    ds.clean.clean_arm_state_variables('testname')
    ds.clean.cleanup()
    ds['qc_atmos_pressure'].attrs['standard_name'] = 'wrong_name'
    ds.clean.link_variables()
    assert ds['qc_atmos_pressure'].attrs['standard_name'] == 'quality_flag'
    ds.close()


def test_qc_flag_description():
    """
    This will check if the cleanup() method will correctly convert convert
    flag_#_description to CF flag_masks and flag_meanings.

    """

    ds = read_arm_netcdf(EXAMPLE_CO2FLX4M)
    ds.clean.cleanup()
    qc_var_name = ds.qcfilter.check_for_ancillary_qc(
        'momentum_flux', add_if_missing=False, cleanup=False
    )

    assert isinstance(ds[qc_var_name].attrs['flag_masks'], list)
    assert isinstance(ds[qc_var_name].attrs['flag_meanings'], list)
    assert isinstance(ds[qc_var_name].attrs['flag_assessments'], list)
    assert ds[qc_var_name].attrs['standard_name'] == 'quality_flag'

    assert len(ds[qc_var_name].attrs['flag_masks']) == 9
    unique_flag_assessments = list({'Acceptable', 'Indeterminate', 'Bad'})
    for f in list(set(ds[qc_var_name].attrs['flag_assessments'])):
        assert f in unique_flag_assessments


def test_clean_sirs_siri_qc():
    ds = read_arm_netcdf(EXAMPLE_SIRS_SIRI_QC)

    data = ds["qc_short_direct_normal"].values
    data[0:5] = 1
    data[5:10] = 2
    data[10:15] = 3
    data[15:20] = 6
    data[20:25] = 7
    data[25:30] = 8
    data[30:35] = 9
    data[35:40] = 94
    data[40:45] = 95
    data[45:50] = 96
    data[50:55] = 97
    data[55:60] = 14
    data[60:65] = 18
    data[65:70] = 22
    data[70:75] = 26
    ds["qc_short_direct_normal"].values = data

    data = ds["qc_up_long_hemisp"].values
    data[0:5] = 1
    data[5:10] = 2
    data[10:15] = 7
    data[15:20] = 8
    data[20:25] = 31
    ds["qc_up_long_hemisp"].values = data

    data = ds["qc_up_short_hemisp"].values
    data[0:5] = 1
    data[5:10] = 2
    data[10:15] = 7
    data[15:20] = 8
    data[20:25] = 31
    ds["qc_up_short_hemisp"].values = data

    ds.clean.cleanup()

    assert ds["qc_short_direct_normal"].attrs['flag_masks'] == [
        1,
        2,
        4,
        8,
        16,
        32,
        64,
        128,
        256,
        512,
        1024,
        2048,
        4096,
        8192,
        16384,
        32768,
    ]
    assert ds["qc_short_direct_normal"].attrs['flag_meanings'] == [
        'Value is set to missing_value.',
        'Passed 1-component test; data fall within max-min limits of Kt,Kn, or Kd',
        'Passed 2-component test; data fall within 0.03 of the Gompertz boundaries',
        'Passed 3-component test; data come within +/- 0.03 of satifying Kt=Kn+Kd',
        'Value estimated; passes all pertinent SERI QC tests',
        'Failed 1-component test; lower than allowed minimum',
        'Falied 1-component test; higher than allowed maximum',
        'Passed 3-component test but failed 2-component test by >0.05',
        'Data fall into a physically impossible region where Kn>Kt by K-space distances of 0.05 to 0.10.',
        'Data fall into a physically impossible region where Kn>Kt by K-space distances of 0.10 to 0.15.',
        'Data fall into a physically impossible region where Kn>Kt by K-space distances of 0.15 to 0.20.',
        'Data fall into a physically impossible region where Kn>Kt by K-space distances of >= 0.20.',
        'Parameter too low by 3-component test (Kt=Kn+Kd)',
        'Parameter too high by 3-component test (Kt=Kn+Kd)',
        'Parameter too low by 2-component test (Gompertz boundary)',
        'Parameter too high by 2-component test (Gompertz boundary)',
    ]

    assert ds["qc_up_long_hemisp"].attrs['flag_masks'] == [1, 2, 4, 8, 16, 32]
    assert ds["qc_up_long_hemisp"].attrs['flag_meanings'] == [
        'Value is set to missing_value.',
        'Passed 1-component test; data fall within max-min limits of up_long_hemisp and down_long_hemisp_shaded, but short_direct_normal and down_short_hemisp or down_short_diffuse fail the SERI QC tests.',
        'Passed 2-component test; data fall within max-min limits of up_long_hemisp and down_long_hemisp_shaded, and short_direct_normal, or down_short_hemisp and down_short_diffuse pass the SERI QC tests while the difference between down_short_hemisp and down_short_diffuse is greater than 20 W/m2.',
        'Failed 1-component test; lower than allowed minimum',
        'Failed 1-component test; higher than allowed maximum',
        'Failed 2-component test',
    ]

    assert ds["qc_up_short_hemisp"].attrs['flag_masks'] == [1, 2, 4, 8, 16, 32]
    assert ds["qc_up_short_hemisp"].attrs['flag_meanings'] == [
        'Value is set to missing_value.',
        'Passed 1-component test',
        'Passed 2-component test',
        'Failed 1-component test; lower than allowed minimum',
        'Failed 1-component test; higher than allowed maximum',
        'Failed 2-component test; solar zenith angle is less than 80 degrees and down_short_hemisp is 0 or missing',
    ]

    assert np.all(ds["qc_short_direct_normal"].values[0:5] == 2)
    assert np.all(ds["qc_short_direct_normal"].values[5:10] == 4)
    assert np.all(ds["qc_short_direct_normal"].values[10:15] == 8)
    assert np.all(ds["qc_short_direct_normal"].values[15:20] == 16)
    assert np.all(ds["qc_short_direct_normal"].values[20:25] == 32)
    assert np.all(ds["qc_short_direct_normal"].values[25:30] == 64)
    assert np.all(ds["qc_short_direct_normal"].values[30:35] == 128)
    assert np.all(ds["qc_short_direct_normal"].values[35:40] == 256)
    assert np.all(ds["qc_short_direct_normal"].values[40:45] == 512)
    assert np.all(ds["qc_short_direct_normal"].values[45:50] == 1024)
    assert np.all(ds["qc_short_direct_normal"].values[50:55] == 2048)
    assert np.all(ds["qc_short_direct_normal"].values[55:60] == 4096)
    assert np.all(ds["qc_short_direct_normal"].values[60:65] == 8192)
    assert np.all(ds["qc_short_direct_normal"].values[65:70] == 16384)
    assert np.all(ds["qc_short_direct_normal"].values[70:75] == 32768)

    assert np.all(ds["qc_up_long_hemisp"].values[0:5] == 2)
    assert np.all(ds["qc_up_long_hemisp"].values[5:10] == 4)
    assert np.all(ds["qc_up_long_hemisp"].values[10:15] == 8)
    assert np.all(ds["qc_up_long_hemisp"].values[15:20] == 16)
    assert np.all(ds["qc_up_long_hemisp"].values[20:25] == 32)

    assert np.all(ds["qc_up_short_hemisp"].values[0:5] == 2)
    assert np.all(ds["qc_up_short_hemisp"].values[5:10] == 4)
    assert np.all(ds["qc_up_short_hemisp"].values[10:15] == 8)
    assert np.all(ds["qc_up_short_hemisp"].values[15:20] == 16)
    assert np.all(ds["qc_up_short_hemisp"].values[20:25] == 32)


def test_swats_qc():
    ds = read_arm_netcdf(EXAMPLE_SWATS)
    ds.clean.cleanup()

    data_var_names = []
    for var_name in ds.data_vars:
        try:
            ds[f'qc_{var_name}']
            data_var_names.append(var_name)
        except KeyError:
            pass

    for var_name in data_var_names:
        qc_var_name = f'qc_{var_name}'

        assert ds[qc_var_name].attrs['flag_masks'] == [1, 2, 4, 8]
        assert ds[qc_var_name].attrs['flag_meanings'] == [
            'Value is set to missing_value.',
            'Data value less than fail_min.',
            'Data value greater than fail_max.',
            'Difference between current and previous values exceeds fail_delta.',
        ]
        assert ds[qc_var_name].attrs['flag_assessments'] == ['Bad', 'Bad', 'Bad', 'Indeterminate']
        assert 'fail_min' in ds[qc_var_name].attrs
        assert 'fail_max' in ds[qc_var_name].attrs
        assert 'fail_delta' in ds[qc_var_name].attrs

        assert 'valid_min' not in ds[var_name].attrs
        assert 'valid_max' not in ds[var_name].attrs
        assert 'valid_delta' not in ds[var_name].attrs
        assert ds[var_name].attrs['units'] != 'C'


def test_fix_incorrect_variable_bit_description_attributes():
    ds = read_arm_netcdf(EXAMPLE_MET_SAIL)
    qc_var_name = 'qc_temp_mean'
    ds[qc_var_name].attrs['qc_bit_2_description'] = ds[qc_var_name].attrs['bit_2_description']
    ds[qc_var_name].attrs['qc_bit_2_assessment'] = ds[qc_var_name].attrs['bit_2_assessment']
    del ds[qc_var_name].attrs['bit_2_description']
    del ds[qc_var_name].attrs['bit_2_assessment']

    ds.clean.cleanup()

    assert ds[qc_var_name].attrs['flag_masks'] == [1, 2, 4, 8]
    assert ds[qc_var_name].attrs['flag_meanings'] == [
        'Value is equal to missing_value.',
        'Value is less than fail_min.',
        'Value is greater than fail_max.',
        'Difference between current and previous values exceeds fail_delta.',
    ]
    assert ds[qc_var_name].attrs['flag_assessments'] == ['Bad', 'Bad', 'Bad', 'Indeterminate']
    assert 'qc_bit_2_description' not in ds[qc_var_name].attrs
    assert 'qc_bit_2_assessment' not in ds[qc_var_name].attrs



"""
Merge multiple datasets
-----------------------

Example to merge multiple data products into one using ACT.
Shows how to adjust the timestamp if the timestamps are at
different part of the sample interval (left, right, center).
Also shows how to apply QC information, merge and resample
data using xarray and plot/write out the results.

"""

from arm_test_data import DATASETS
import matplotlib.pyplot as plt
import xarray as xr

import act

# Set data files
# An alternative to this is to download data from the
# ARM Data Webservice as shown in the discovery plot_neon.py example
ebbr_file = DATASETS.fetch('sgp30ebbrE13.b1.20190601.000000.nc')
ecor_file = DATASETS.fetch('sgp30ecorE14.b1.20190601.000000.cdf')
sebs_file = DATASETS.fetch('sgpsebsE14.b1.20190601.000000.cdf')

# Read data into datasets
ds_ebbr = act.io.arm.read_arm_netcdf(ebbr_file)
ds_ecor = act.io.arm.read_arm_netcdf(ecor_file)
ds_sebs = act.io.arm.read_arm_netcdf(sebs_file)

# Check for ARM DQRs and add them to the QC variables
ds_ebbr = act.qc.arm.add_dqr_to_qc(ds_ebbr)
ds_ecor = act.qc.arm.add_dqr_to_qc(ds_ecor)
ds_sebs = act.qc.arm.add_dqr_to_qc(ds_sebs)

# The ECOR and EBBR have different definitions of latent heat
# flux and what is positive vs negative.  Check out the ARM
# Handbooks for more information
ds_ecor['lv_e'].values = ds_ecor['lv_e'].values * -1.0

# For example purposes, let's rename the ecor latent heat flux
ds_ecor = ds_ecor.rename({'lv_e': 'latent_heat_flux_ecor'})
ds_ecor['latent_heat_flux_ecor'].attrs['ancillary_variables'] = 'qc_latent_heat_flux_ecor'
ds_ecor = ds_ecor.rename({'qc_lv_e': 'qc_latent_heat_flux_ecor'})

# Also going to Switch some QC for example purposes
qc = ds_ecor['qc_latent_heat_flux_ecor'].values
qc[10:20] = 2
ds_ecor['qc_latent_heat_flux_ecor'].values = qc

# There is a difference in how these timestamps are defined
# The EBBR is at the end of the sampling interval and the
# ECOR is at the beginning.  Knowing this, we can shift the
# EBBR timestampes by 30 minutes to coincide with the ECOR
ds_ebbr = act.utils.datetime_utils.adjust_timestamp(ds_ebbr, offset=-30 * 60)

# Now, we can merge all these datasets into one product
ds = xr.merge([ds_ecor, ds_ebbr, ds_sebs], compat='override')

# Apply the QC information to set all flagged data to missing/NaN
ds.qcfilter.datafilter(
    del_qc_var=False, rm_assessments=['Bad', 'Incorrect', 'Indeterminate', 'Suspect']
)

# Plot up data from the merged dataset for each of the instruments
display = act.plotting.TimeSeriesDisplay(ds, figsize=(15, 10), subplot_shape=(3,))
display.plot('latent_heat_flux_ecor', label='ECOR', subplot_index=(0,))
display.plot('latent_heat_flux', label='EBBR', subplot_index=(0,))
plt.legend()
display.plot('surface_soil_heat_flux_1', label='SEBS', subplot_index=(1,))

# Plot out the QC information that was modified as well
display.qc_flag_block_plot('latent_heat_flux_ecor', subplot_index=(2,))
plt.show()

# Resample the data to 1 hour mean
# Check out the xarray documentation for more information
# on the resample function. Options include mean, median,
# max, min, sum, nearest, and more.
ds = ds.resample(time='H').mean(keep_attrs=True)

# Plot up data from the hourly merged dataset for ECOR and EBBR
display = act.plotting.TimeSeriesDisplay(ds, figsize=(15, 10), subplot_shape=(1,))
display.plot('latent_heat_flux_ecor', label='ECOR', subplot_index=(0,))
display.plot('latent_heat_flux', label='EBBR', subplot_index=(0,))
plt.show()

# Write data out to netcdf
ds.to_netcdf('./sgpecor_ebbr_sebs.nc')



"""
Script for downloading data from the IMPROVE network

"""

import pandas as pd
import numpy as np
import xarray as xr


def get_improve_data(site_id=None, parameter_id=None, start_date=None, end_date=None):
    """
    Retrieve IMPROVE data for the given site and variable ids and store it in an
    xarray dataset. Documentation on the IMPROVE data can be found at
    https://vista.cira.colostate.edu/Improve/data-user-guide/

    Also adds in metadata from the site summary page to the global attributes
    https://views.cira.colostate.edu/adms/Pub/SiteSummary.aspx?dsidse=10001&siidse=244

    Parameters
    ----------
    site_id : str
        Site id number which can be retrieved from the IMPROVE page for each site such as
        https://views.cira.colostate.edu/adms/Pub/SiteSummary.aspx?dsidse=10001&siidse=244
    parameter_id : list
        List of parameter id values to retrieve from the API.
    start_date : str
        Start date formatted as M/D/YEAR such as 1/31/2022
    end_date : str
        End date formatted as M/D/YEAR such as 1/31/2022

    Returns
    -------
    ds : xarray.Dataset
        Returns an Xarray dataset object

    Example
    -------
    act.discovery.get_improve_data(site_id='244')

    """

    # Build URL
    base_url = 'https://views.cira.colostate.edu/fed/svc/DataSvc.aspx?action=getqueryresults&cmdfileid=ServiceSqlCommandFile&cmdid=BasicDataQuery1_Codes'

    if site_id is None:
        raise ValueError('Please provide a site_id')
    else:
        base_url += '&dsidse=10001&siidse=' + str(site_id)

    if parameter_id is None:
        base_url += '&paidse=101,136,907,900,102,104,105,115,116,117,114,3778,142,143,144,145,3016,146,3699,141,3779,3217,108,109,112,113,301,304,303,3716,3717,3718,3719,3720,3721,3722,3730,3731,3732,3733,3734,3735,3736,3694,121,3723,3724,3725,3726,3727,3728,3729,3737,3738,3739,3740,3741,3742,3743,118,148,128,130,132,941,127,903,910,3744,3745,3746,3747,3748,3749,3750,3751,3752,3753,3754,3755,3756,3757,131,138,139,133,3704,3705,3706,3707,3708,3709,3710,3711,3712,3713,3714,3715,147,124,150,3695,3014,153,154,134,911,158,156,151,202,159,160,162,163'
    else:
        base_url += '&paidse=' + ','.join(parameter_id)

    if start_date is None:
        raise ValueError('Please provide a start date')
    else:
        base_url += '&sd=' + start_date
    if end_date is None:
        raise ValueError('Please provide an end date')
    else:
        base_url += '&ed=' + end_date

    # Read data and get variables
    df = pd.read_html(base_url)[0]
    variables = np.unique(df.Param)

    # Print out proper acknowledgement
    print("Please use the following acknowledgment when using IMPROVE data:\n")

    print(
        "IMPROVE is a collaborative association of state, tribal, and federal agencies, and international partners. US Environmental Protection Agency is the primary funding source, with contracting and research support from the National Park Service. The Air Quality Group at the University of California, Davis is the central analytical laboratory, with ion analysis provided by Research Triangle Institute, and carbon analysis provided by Desert Research Institute."
    )

    # Creat mapping of variable names to metadata
    mapping = {
        'ALf': {'name': 'aluminum_fine', 'long_name': 'Aluminum (Fine)', 'epa_code': '88104'},
        'ASf': {'name': 'arsenic_fine', 'long_name': 'Arsenic (Fine)', 'epa_code': '88103'},
        'BRf': {'name': 'bromine_fine', 'long_name': 'Bromine (Fine)', 'epa_code': '88109'},
        'CAf': {'name': 'calcium_fine', 'long_name': 'Calcium (Fine)', 'epa_code': '88111'},
        'CLf': {'name': 'chlorine_fine', 'long_name': 'Chlorine (Fine)', 'epa_code': '88115'},
        'CRf': {'name': 'chromium_fine', 'long_name': 'Chromium (Fine)', 'epa_code': '88112'},
        'CUf': {'name': 'copper_fine', 'long_name': 'Copper (Fine)', 'epa_code': '88114'},
        'FEf': {'name': 'iron_fine', 'long_name': 'Iron (Fine)', 'epa_code': '88126'},
        'PBf': {'name': 'lead_fine', 'long_name': 'Lead (Fine)', 'epa_code': '88128'},
        'MGf': {'name': 'magnesium_fine', 'long_name': 'Magnesium (Fine)', 'epa_code': '88140'},
        'MNf': {'name': 'manganese_fine', 'long_name': 'Manganese (Fine)', 'epa_code': '88132'},
        'NIf': {'name': 'nickel_fine', 'long_name': 'Nickel (Fine)', 'epa_code': '88136'},
        'Pf': {'name': 'phosphorus_fine', 'long_name': 'Phosphorus (Fine)', 'epa_code': '88152'},
        'Kf': {'name': 'potassium_fine', 'long_name': 'Potassium (Fine)', 'epa_code': '88180'},
        'RBf': {'name': 'rubidium_fine', 'long_name': 'Rubidium (Fine)', 'epa_code': '88176'},
        'SEf': {'name': 'selenium_fine', 'long_name': 'Selenium (Fine)', 'epa_code': '88154'},
        'SIf': {'name': 'silicon_fine', 'long_name': 'Silicon (Fine)', 'epa_code': '88165'},
        'NAf': {'name': 'sodium_fine', 'long_name': 'Sodium (Fine)', 'epa_code': '88184'},
        'SRf': {'name': 'strontium_fine', 'long_name': 'Strontium (Fine)', 'epa_code': '88168'},
        'Sf': {'name': 'sulfur_fine', 'long_name': 'Sulfur (Fine)', 'epa_code': '88169'},
        'TIf': {'name': 'titanium_fine', 'long_name': 'Titanium (Fine)', 'epa_code': '88161'},
        'Vf': {'name': 'vanadium_fine', 'long_name': 'Vanadium (Fine)', 'epa_code': '88164'},
        'ZNf': {'name': 'zinc_fine', 'long_name': 'Zinc (Fine)', 'epa_code': '88167'},
        'ZRf': {'name': 'zirconium_fine', 'long_name': 'Zirconium (Fine)', 'epa_code': '88185'},
        'CHLf': {'name': 'chloride_fine', 'long_name': 'Chloride (Fine)', 'epa_code': '88203'},
        'NO3f': {'name': 'nitrate_fine', 'long_name': 'Nitrate (Fine)', 'epa_code': '88306'},
        'N2f': {'name': 'nitrite_fine', 'long_name': 'Nitrite (Fine)', 'epa_code': '88338'},
        'SO4f': {'name': 'sulfate_fine', 'long_name': 'Sulfate (Fine)', 'epa_code': '88403'},
        'OC1f': {
            'name': 'carbon_organic_fraction_1_fine',
            'long_name': 'Carbon, Organic Fraction 1 (Fine)',
            'comments': 'TOR, pure helium (>99.999%) atmosphere, temperature (T) = 140 °C',
            'epa_code': '88324',
        },
        'OC2f': {
            'name': 'carbon_organic_fraction_2_fine',
            'long_name': 'Carbon, Organic Fraction 2 (Fine)',
            'comments': 'TOR, pure helium (>99.999%) atmosphere, temperature (T) = 280 °C',
            'epa_code': '88325',
        },
        'OC3f': {
            'name': 'carbon_organic_fraction_3_fine',
            'long_name': 'Carbon, Organic Fraction 3 (Fine)',
            'comments': 'TOR, pure helium (>99.999%) atmosphere, temperature (T) = 480 °C',
            'epa_code': '88326',
        },
        'OC4f': {
            'name': 'carbon_organic_fraction_4_fine',
            'long_name': 'Carbon, Organic Fraction 4 (Fine)',
            'comments': 'TOR, pure helium (>99.999%) atmosphere, temperature (T) = 580 °C',
            'epa_code': '88327',
        },
        'OPf': {
            'name': 'carbon_organic_reflectance_fine',
            'long_name': 'Carbon, Organic Pyrolized (Fine) by Reflectance',
            'comments': 'TOR, carbon that is measured after the introduction of helium/oxygen atmosphere at °550 C but beforereflectance returns to initial value',
            'epa_code': '88328',
        },
        'OPTf': {
            'name': 'carbon_organic_transmittance_fine',
            'long_name': 'Carbon, Organic Pyrolized (Fine) by Transmittance',
            'comments': 'TOR, carbon that is measured after the introduction of helium/oxygen atmosphere at °550 C but beforetransmittance returns to initial value',
            'epa_code': '88336',
        },
        'OCf': {
            'name': 'carbon_organic_total_fine',
            'long_name': 'Carbon, Organic Total (Fine)',
            'comments': 'Organic carbon from TOR carbon fractions (OC1f+OC2f+OC3f+OC4f+OPf)',
            'epa_code': '88320',
        },
        'EC1f': {
            'name': 'carbon_elemental_fraction_1_fine',
            'long_name': 'Carbon, Elemental Fraction 1 (Fine)',
            'comments': 'TOR, 98% helium, 2% oxygen atmosphere, temperature (T) = 580° C.',
            'epa_code': '88329',
        },
        'EC2f': {
            'name': 'carbon_elemental_fraction_2_fine',
            'long_name': 'Carbon, Elemental Fraction 2 (Fine)',
            'comments': 'TOR, 98% helium, 2% oxygen atmosphere, temperature (T) = 740° C.',
            'epa_code': '88380',
        },
        'EC3f': {
            'name': 'carbon_elemental_fraction_3_fine',
            'long_name': 'Carbon, Elemental Fraction 3 (Fine)',
            'comments': 'TOR, 98% helium, 2% oxygen atmosphere, temperature (T) = 840° C.',
            'epa_code': '88331',
        },
        'ECf': {
            'name': 'carbon_elemental_total_fine',
            'long_name': 'Carbon, Elemental Total (Fine)',
            'comments': 'Elemental carbon from TOR carbon fractions (E1+E2+E3-OP)',
            'epa_code': '88321',
        },
        'fAbs': {
            'name': 'filter_absorption_coeff',
            'long_name': 'Filter Absorption Coefficient',
            'comments': 'A calibrated absorption coefficient measured from a Teflon filter using a hybrid integrating plate and sphere (HIPS) method',
            'epa_code': '63102',
        },
        'FlowRate': {
            'name': 'flow_rate',
            'long_name': 'Flow Rate',
            'comments': 'The rate of air flow through an air sampling instrument',
            'epa_code': '63102',
        },
        'MF': {
            'name': 'mass_pm2_5',
            'long_name': 'Mass, PM2.5 (Fine)',
            'comments': 'Gravimetric mass measurement for particles with aerodynamic diameters less than 2.5 um',
            'epa_code': '88101',
        },
        'MT': {
            'name': 'mass_pm10',
            'long_name': 'Mass, PM10 (Total)',
            'comments': 'Gravimetric mass measurement for particles with aerodynamic diameters less than 10 um',
            'epa_code': '85101',
        },
        'SampDur': {
            'name': 'sample_duration',
            'long_name': 'Sampling Duration',
            'comments': 'The duration of a given sampling period in minutes',
        },
        'ammNO3f': {
            'name': 'ammonium_nitrate_fine',
            'long_name': 'Ammonium Nitrate (Fine)',
            'comments': '1.29 x  NO3f',
        },
        'ammSO4f': {
            'name': 'ammonium_sulfate_fine',
            'long_name': 'Ammonium Sulfate (Fine)',
            'comments': '1.375 x SO4f',
        },
        'OMCf': {
            'name': 'carbon_organic_mass_fine',
            'long_name': 'Carbon, Organic Mass (fine)(1.8*OC)',
            'comments': '1.8 X OCf',
        },
        'TCf': {
            'name': 'carbon_total_fine',
            'long_name': 'Carbon, Total (fine)',
            'comments': 'From TOR carbon fractions (OCf+ECf)',
        },
        'CM_calculated': {
            'name': 'CM_calculated',
            'long_name': 'Mass, PM10-PM2.5 (Coarse)',
            'comments': 'MT-MF',
        },
        'SeaSaltf': {
            'name': 'sea_salt_fine',
            'long_name': 'Sea Salt (Fine)',
            'comments': '1.8XCHLf',
        },
        'SOILf': {
            'name': 'soil_fine',
            'long_name': 'Soil (Fine)',
            'comments': '2.2 × ALf + 2.49 × SIf + 1.63 × CAf + 2.42 × FEf + 1.94 × TIf',
        },
        'RCFM': {
            'name': 'mass_pm2_5_reconstructed',
            'long_name': 'Mass, PM2.5 Reconstructed (Fine)',
            'comments': 'Sum of ammSO4f, ammNO3f, OMCf, ECf, soilf, and seasaltf.',
        },
        'RCTM': {
            'name': 'mass_pmi10_reconstructed',
            'long_name': 'Mass, PM10 Reconstructed (Total)',
            'comments': 'Sum of ammSO4f, ammNO3f, OMCf, ECf, soilf, seasaltf, and CM_calculated.',
        },
    }
    laser_vars = {
        'RefF': {
            'units': 'ratio',
            'comments': 'Final laser reflectance at ',
            'name': 'final_laser_reflectance_',
        },
        'TransF': {
            'units': 'ratio',
            'comments': 'Final laser transmittance at ',
            'name': 'final_laser_transmittance_',
        },
        'RefI': {
            'units': 'ratio',
            'comments': 'Initial laser reflectance at ',
            'name': 'initial_laser_reflectance_',
        },
        'TransI': {
            'units': 'ratio',
            'comments': 'Initial laser transmittance at ',
            'name': 'initial_laser_transmittance_',
        },
        'RefM': {
            'units': 'ratio',
            'comments': 'Minimum laser reflectance at ',
            'name': 'min_laser_reflectance_',
        },
        'TransM': {
            'units': 'ratio',
            'comments': ' Minimum laser transmittance at ',
            'name': 'min_laser_transmittance_',
        },
        'OP_TR': {
            'units': 'ug m-3',
            'comments': 'Organic Pyrolyzed Carbon by Reflectance at ',
            'name': 'organic_pyrolyzed_carbon_reflectance_',
        },
        'OP_TT': {
            'units': 'ug m-3',
            'comments': 'Organic Pyrolyzed Carbon by Transmittance at ',
            'name': 'organic_pyrolyzed_carbon_transmittance_',
        },
    }
    laser_wl = ['405', '445', '532', '635', '780', '808', '980']
    for v in laser_vars:
        for wl in laser_wl:
            name = laser_vars[v]['name'] + wl
            if 'OP' not in v:
                mapping['_'.join([v, wl])] = {
                    'units': laser_vars[v]['units'],
                    'name': name,
                    'long_name': ' '.join([laser_vars[v]['comments'], wl]),
                }
            else:
                var_name = wl.join(v.split('_'))
                mapping[var_name] = {
                    'units': laser_vars[v]['units'],
                    'name': name,
                    'long_name': ' '.join([laser_vars[v]['comments'], wl]),
                }

    # Run through each variable in the dataframe and add it to a dataset
    # along with the appropriate metadata
    ct = 0
    site = np.unique(df.Site)[0]
    attrs = {'url': base_url, 'datastream': site + ' IMPROVE'}

    for v in variables:
        # Find data for just the variable in question
        poc_attrs = {'units': '1', 'long_name': 'Parameter Occurrence Code for ' + v}
        df2 = df[df.Param == v]

        # Get metadata
        unit = np.unique(df2.UnitAbbr)
        if len(unit) > 1:
            raise ValueError('Multiple types of units detected, using first one')

        sites = np.unique(df2.Site)
        if len(sites) > 1:
            raise ValueError('Multiple sites detected, please use only one')

        # Get time, POC, and data
        time = pd.to_datetime(df2.FactDate)
        poc = df2.POC

        data = df2.FactValue

        # Set up attributes
        var_attrs = {'units': unit[0], 'long_name': mapping[v]['long_name'], '_FillValue': -999.0}
        if 'comments' in mapping[v]:
            var_attrs['comments'] = mapping[v]['comments']
        if 'epa_code' in mapping[v]:
            var_attrs['epa_code'] = mapping[v]['epa_code']

        # If the first variable, create the dataset and then add variables to it
        if ct == 0:
            ds = xr.Dataset(
                data_vars={mapping[v]['name']: (['time'], data, var_attrs)},
                coords={'time': time},
                attrs=attrs,
            )
            ds['poc_' + mapping[v]['name']] = xr.DataArray(
                data=poc, dims=['time'], coords={'time': time}, attrs=poc_attrs
            )
            ct += 1
        else:
            ds[mapping[v]['name']] = xr.DataArray(
                data=data, dims=['time'], coords={'time': time}, attrs=var_attrs
            )
            ds['poc_' + mapping[v]['name']] = xr.DataArray(
                data=poc, dims=['time'], coords={'time': time}, attrs=poc_attrs
            )

    # Add in metadata from site summary page
    url = 'https://views.cira.colostate.edu/adms/Pub/SiteSummary.aspx?dsidse=10001&siidse=' + str(
        site_id
    )
    df = pd.read_html(url)
    for i in df[0].index:
        # Add lat/lon as variables
        if df[0][0][i] == 'Latitude':
            attrs = {
                'long_name': 'North latitude',
                'units': 'degree_N',
                'valid_min': -90.0,
                'valid_max': 90.0,
                'standard_name': 'latitude',
            }
            ds['lat'] = xr.DataArray(
                data=float(df[0][1][i]),
                dims=['time'],
                coords={'time': ds['time'].values},
                attrs=attrs,
            )
        elif df[0][0][i] == 'Longitude':
            attrs = {
                'long_name': 'East longitude',
                'units': 'degree_E',
                'valid_min': -180.0,
                'valid_max': 180.0,
                'standard_name': 'longitude',
            }
            ds['lon'] = xr.DataArray(
                data=float(df[0][1][i]),
                dims=['time'],
                coords={'time': ds['time'].values},
                attrs=attrs,
            )
        else:
            ds.attrs[df[0][0][i]] = df[0][1][i]

    # Add in problem information from the site summary page
    problem = ''
    for i in df[-1].index:
        problem += '_'.join(
            [df[-1]['EventDate'][i], df[-1]['EventType'][i], df[-1]['Notes'][i], '\n']
        )
    ds.attrs['site_problems'] = problem

    ds = ds.sortby('time')

    return ds



"""
Changing units in dataset
-------------------------

This is an example of how to change
units in the xarray dataset.

"""

from arm_test_data import DATASETS
import numpy as np

import act


def print_summary(ds, variables):
    for var_name in variables:
        print(
            f'{var_name}: mean={np.nanmean(ds[var_name].values)} '
            f"units={ds[var_name].attrs['units']}"
        )
    print()


variables = ['first_cbh', 'second_cbh', 'alt']

# Read in some example data
filename_ceil = DATASETS.fetch('sgpceilC1.b1.20190101.000000.nc')
ds = act.io.arm.read_arm_netcdf(filename_ceil)

# Print the variable name, mean of values and units
print('Variables in read data')
print_summary(ds, variables)

# Change units of one varible from m to km
ds.utils.change_units(variables='first_cbh', desired_unit='km')
print('Variables with one changed to km')
print_summary(ds, variables)

# Change units of more than one varible from to km
ds.utils.change_units(variables=variables, desired_unit='km')
print('Variables with both changed to km')
print_summary(ds, variables)

# Can change all data variables in the dataset that are units of length by not providing
# a list of variables. Here we are changing back to orginal meters.
# Also, because it needs to loop over all variables and try to convert, will take
# longer if we keep the QC variables. Faseter if we exclude them.
# The method will return a dataset. In this case the dataset returned is the same
# dataset.
skip_variables = [ii for ii in ds.data_vars if ii.startswith('qc_')]
new_ds = ds.utils.change_units(variables=None, desired_unit='m', skip_variables=skip_variables)
print('Variables changed back to m by looping over all variables in dataset')
print('Orginal dataset is same as retured dataset:', ds is new_ds)
print_summary(new_ds, variables)

# For coordinate variables need to explicitly give coordinage variable name and use
# the returned dataset. The xarray method used to change values on coordinate
# values requries returning a new updated dataet.
var_name = 'range'
variables.append(var_name)
new_ds = ds.utils.change_units(variables=variables, desired_unit='km')
print('Variables and coordinate variable values changed to km')
print('Orginal dataset is same as retured dataset:', ds is new_ds)
print_summary(new_ds, variables)



"""
NASA MPLNET
-----------

This example shows how to download data from
NASA's MicroPulsed Lidar Network

"""

import act

# Determine MPLNET site of interest
meta = act.discovery.get_mplnet_meta(sites="active", method="data", print_to_screen=True)

# Download MPLNET data for site of interest
output = act.discovery.download_mplnet_data(
    version=3, level=1, product="NRB", site="GSFC", year="2022", month="09", day="01"
)



"""
NEON Data
---------

This example shows how to download data from
NEON and ARM 2m surface meteorology stations
on the North Slope and plot them

"""

import os
import glob
import matplotlib.pyplot as plt

import act

# Place your username and token here
username = os.getenv('ARM_USERNAME')
token = os.getenv('ARM_PASSWORD')

if token is not None and len(token) > 0:
    # Download ARM data if a username/token are set
    files = act.discovery.download_arm_data(
        username, token, 'nsametC1.b1', '2022-10-01', '2022-10-07'
    )
    ds = act.io.arm.read_arm_netcdf(files)

    # Download NEON Data
    # NEON sites can be found through the NEON website
    # https://www.neonscience.org/field-sites/explore-field-sites
    site_code = 'BARR'
    product_code = 'DP1.00002.001'
    result = act.discovery.neon.download_neon_data(site_code, product_code, '2022-10')

    # A number of files are downloaded and further explained in the readme file that's downloaded.
    # These are the files we will need for reading 1 minute NEON data
    file = glob.glob(
        os.path.join(
            '.',
            'BARR_DP1.00002.001',
            'NEON.D18.BARR.DP1.00002.001.000.010.001.SAAT_1min.2022-10.expanded.*.csv',
        )
    )
    variable_file = glob.glob(
        os.path.join('.', 'BARR_DP1.00002.001', 'NEON.D18.BARR.DP1.00002.001.variables.*.csv')
    )
    position_file = glob.glob(
        os.path.join(
            '.',
            'BARR_DP1.00002.001',
            'NEON.D18.BARR.DP1.00002.001.sensor_positions.*.csv',
        )
    )
    # Read in the data using the ACT reader, passing with it the variable and position files
    # for added information in the dataset
    ds2 = act.io.read_neon_csv(file, variable_files=variable_file, position_files=position_file)

    # Plot up the two datasets
    display = act.plotting.TimeSeriesDisplay({'ARM': ds, 'NEON': ds2})
    display.plot('temp_mean', 'ARM', marker=None, label='ARM')
    display.plot('tempSingleMean', 'NEON', marker=None, label='NEON')
    display.day_night_background('ARM')
    plt.show()



import glob
import os

import numpy as np
import pytest

import act


def test_download_armdata():
    if not os.path.isdir(os.getcwd() + '/data/'):
        os.makedirs(os.getcwd() + '/data/')

    # Place your username and token here
    username = os.getenv('ARM_USERNAME')
    token = os.getenv('ARM_PASSWORD')

    if username is not None and token is not None:
        if len(username) == 0 and len(token) == 0:
            return
        datastream = 'sgpmetE13.b1'
        startdate = '2020-01-01'
        enddate = startdate
        outdir = os.getcwd() + '/data/'

        results = act.discovery.arm.download_arm_data(
            username, token, datastream, startdate, enddate, output=outdir
        )
        files = glob.glob(outdir + datastream + '*20200101*cdf')
        if len(results) > 0:
            assert files is not None
            assert 'sgpmetE13' in files[0]

        if files is not None:
            if len(files) > 0:
                os.remove(files[0])

        datastream = 'sgpmeetE13.b1'
        act.discovery.arm.download_arm_data(
            username, token, datastream, startdate, enddate, output=outdir
        )
        files = glob.glob(outdir + datastream + '*20200101*cdf')
        assert len(files) == 0

        with np.testing.assert_raises(ConnectionRefusedError):
            act.discovery.arm.download_arm_data(
                username, token + '1234', datastream, startdate, enddate, output=outdir
            )

        datastream = 'sgpmetE13.b1'
        results = act.discovery.arm.download_arm_data(
            username, token, datastream, startdate, enddate
        )
        assert len(results) == 1


def test_download_armdata_hourly():
    if not os.path.isdir(os.getcwd() + '/data/'):
        os.makedirs(os.getcwd() + '/data/')

    # Place your username and token here
    username = os.getenv('ARM_USERNAME')
    token = os.getenv('ARM_PASSWORD')

    if username is not None and token is not None:
        if len(username) == 0 and len(token) == 0:
            return
        datastream = 'sgpmetE13.b1'
        startdate = '2020-01-01T00:00:00'
        enddate = '2020-01-01T12:00:00'
        outdir = os.getcwd() + '/data/'

        results = act.discovery.arm.download_arm_data(
            username, token, datastream, startdate, enddate, output=outdir
        )
        files = glob.glob(outdir + datastream + '*20200101*cdf')
        if len(results) > 0:
            assert files is not None
            assert 'sgpmetE13' in files[0]

        if files is not None:
            if len(files) > 0:
                os.remove(files[0])

        datastream = 'sgpmeetE13.b1'
        act.discovery.arm.download_arm_data(
            username, token, datastream, startdate, enddate, output=outdir
        )
        files = glob.glob(outdir + datastream + '*20200101*cdf')
        assert len(files) == 0

        with np.testing.assert_raises(ConnectionRefusedError):
            act.discovery.arm.download_arm_data(
                username, token + '1234', datastream, startdate, enddate, output=outdir
            )

        datastream = 'sgpmetE13.b1'
        results = act.discovery.arm.download_arm_data(
            username, token, datastream, startdate, enddate
        )
        assert len(results) == 1


def test_download_arm_data_mod_with_variables():
    # Call the function
    username = os.getenv('ARM_USERNAME')
    token = os.getenv('ARM_PASSWORD')
    if username is not None and token is not None:
        if len(username) == 0 and len(token) == 0:
            return

        file_list = [
            "sgpmetE13.b1.20200101.000000.cdf",
            "sgpmetE13.b1.20200102.000000.cdf",
        ]
        variables = [
            "temp_mean",
            "rh_mean",
        ]

        act.discovery.download_arm_data_mod(username, token, file_list, variables=variables)
        # Check if the file was saved correctly
        assert os.path.exists('sgpmetE13.b1.20200101-20200102.cdf')

        # Check if variables are only what was selected
        ds = act.io.read_arm_netcdf('sgpmetE13.b1.20200101-20200102.cdf')
        assert len(ds.variables.keys()) == 3

        # Check if all variables are present
        act.discovery.download_arm_data_mod(username, token, file_list)
        ds = act.io.read_arm_netcdf('sgpmetE13.b1.20200101-20200102.cdf')
        assert len(ds.variables.keys()) == 50

        # Check if custom output path exist and directory is created.
        act.discovery.download_arm_data_mod(
            username, token, file_list, output=os.getcwd() + '/data/'
        )
        ds = act.io.read_arm_netcdf(os.getcwd() + '/data/' + 'sgpmetE13.b1.20200101-20200102.cdf')
        assert len(ds.variables.keys()) == 50


def test_download_arm_data_mod_csv():
    username = os.getenv('ARM_USERNAME')
    token = os.getenv('ARM_PASSWORD')
    if username is not None and token is not None:
        if len(username) == 0 and len(token) == 0:
            return

        file_list = [
            "sgpmetE13.b1.20200101.000000.cdf",
            "sgpmetE13.b1.20200102.000000.cdf",
        ]
        # Test if csv file is created
        act.discovery.download_arm_data_mod(username, token, file_list, filetype='csv')

        # Check if the file was saved correctly
        assert os.path.exists('sgpmetE13.b1.20200101-20200102.csv')


def test_download_arm_data_mod_warn():
    # Check for warning if parameters are wrong
    with pytest.warns(UserWarning):
        username = 'foo'
        token = 'bar'
        file_list = [
            "sgpmetE13.b1.20200101.000000.cdf",
            "sgpmetE13.b1.20200102.000000.cdf",
        ]
        act.discovery.download_arm_data_mod(username, token, file_list, filetype='csv')


def test_arm_doi():
    datastream = 'sgpmetE13.b1'
    startdate = '2022-01-01'
    enddate = '2022-12-31'
    doi = act.discovery.get_arm_doi(datastream, startdate, enddate)

    assert len(doi) > 10
    assert isinstance(doi, str)
    assert 'doi' in doi
    assert 'Kyrouac' in doi

    doi = act.discovery.get_arm_doi('test', startdate, enddate)
    assert 'No DOI Found' in doi



"""
Convert Data to AmeriFlux Format
--------------------------------

This script shows how to convert ARM data to AmeriFlux format
using an ACT function, and write it out to csv.  More information
on AmeriFlux and their file formats and naming conventions can be
found here: https://ameriflux.lbl.gov/

Author: Adam Theisen

"""

import act
import glob
import xarray as xr
import os
import matplotlib.pyplot as plt

# Read in the ECOR data
files = glob.glob(act.tests.sample_files.EXAMPLE_ECORSF_E39)
ds_ecor = act.io.arm.read_arm_netcdf(files)

# The ECOR time stamp as at the end of the Averaging period so adjusting
# it to be consistent with the other systems
ds_ecor = act.utils.datetime_utils.adjust_timestamp(ds_ecor)

# Clean up and QC the data based on embedded QC and ARM DQRs
ds_ecor.clean.cleanup()
ds_ecor = act.qc.arm.add_dqr_to_qc(ds_ecor)
ds_ecor.qcfilter.datafilter(
    del_qc_var=False, rm_assessments=['Bad', 'Incorrect', 'Indeterminate', 'Suspect']
)

# Then we do this same thing for the other instruments
# SEBS
files = glob.glob(act.tests.sample_files.EXAMPLE_SEBS_E39)
ds_sebs = act.io.arm.read_arm_netcdf(files)
# SEBS does not have a time_bounds variable so we have to manually adjust it
ds_sebs = act.utils.datetime_utils.adjust_timestamp(ds_sebs, offset=-30 * 60)
ds_sebs.clean.cleanup()
ds_sebs = act.qc.arm.add_dqr_to_qc(ds_sebs)
ds_sebs.qcfilter.datafilter(
    del_qc_var=False, rm_assessments=['Bad', 'Incorrect', 'Indeterminate', 'Suspect']
)

# STAMP
files = glob.glob(act.tests.sample_files.EXAMPLE_STAMP_E39)
ds_stamp = act.io.arm.read_arm_netcdf(files)
ds_stamp.clean.cleanup()
ds_stamp = act.qc.arm.add_dqr_to_qc(ds_stamp)
ds_stamp.qcfilter.datafilter(
    del_qc_var=False, rm_assessments=['Bad', 'Incorrect', 'Indeterminate', 'Suspect']
)

# STAMP Precipitation
files = glob.glob(act.tests.sample_files.EXAMPLE_STAMPPCP_E39)
ds_stamppcp = act.io.arm.read_arm_netcdf(files)
ds_stamppcp.clean.cleanup()
ds_stamppcp = act.qc.arm.add_dqr_to_qc(ds_stamppcp)
ds_stamppcp.qcfilter.datafilter(
    del_qc_var=False, rm_assessments=['Bad', 'Incorrect', 'Indeterminate', 'Suspect']
)
# These are minute data so we need to resample and sum up to 30 minutes
ds_stamppcp = ds_stamppcp['precip'].resample(time='30Min').sum()

# AMC
files = glob.glob(act.tests.sample_files.EXAMPLE_AMC_E39)
ds_amc = act.io.arm.read_arm_netcdf(files)
ds_amc.clean.cleanup()
ds_amc = act.qc.arm.add_dqr_to_qc(ds_amc)
ds_amc.qcfilter.datafilter(
    del_qc_var=False, rm_assessments=['Bad', 'Incorrect', 'Indeterminate', 'Suspect']
)

# Merge these datasets together
ds = xr.merge([ds_ecor, ds_sebs, ds_stamp, ds_stamppcp, ds_amc], compat='override')

# Convert the data to AmeriFlux format and get a DataFrame in return
# Note, this does not return an xarray Dataset as it's assumed the data
# will just be written out to csv format.
df = act.io.ameriflux.convert_to_ameriflux(ds)

# Write the data out to file
site = 'US-A14'
directory = './' + site + 'mergedflux/'
if not os.path.exists(directory):
    os.makedirs(directory)

# Following the AmeriFlux file naming convention
filename = (
    site
    + '_HH_'
    + str(df['TIMESTAMP_START'].iloc[0])
    + '_'
    + str(df['TIMESTAMP_END'].iloc[-1])
    + '.csv'
)
df.to_csv(directory + filename, index=False)


# Plot up merged data for visualization
display = act.plotting.TimeSeriesDisplay(ds, subplot_shape=(4,), figsize=(12, 10))
display.plot('latent_flux', subplot_index=(0,))
display.plot('co2_flux', subplot_index=(0,))
display.plot('sensible_heat_flux', subplot_index=(0,))
display.day_night_background(subplot_index=(0,))

display.plot('precip', subplot_index=(1,))
display.day_night_background(subplot_index=(1,))

display.plot('surface_soil_heat_flux_1', subplot_index=(2,))
display.plot('surface_soil_heat_flux_2', subplot_index=(2,))
display.plot('surface_soil_heat_flux_3', subplot_index=(2,))
display.day_night_background(subplot_index=(2,))

display.plot('soil_specific_water_content_west', subplot_index=(3,))
display.axes[3].set_ylim(display.axes[3].get_ylim()[::-1])

display.day_night_background(subplot_index=(3,))

plt.subplots_adjust(hspace=0.35)
plt.show()



import random
import shutil
import tempfile
from os import PathLike, chdir
from pathlib import Path
from string import ascii_letters

import numpy as np
import pytest

import act
from act.tests import sample_files

try:
    import moviepy.video.io.ImageSequenceClip  # noqa

    MOVIEPY_AVAILABLE = True
except (ImportError, RuntimeError):
    MOVIEPY_AVAILABLE = False


def test_read_netcdf_gztarfiles():
    with tempfile.TemporaryDirectory() as tmpdirname:
        met_files = list(Path(file) for file in act.tests.EXAMPLE_MET_WILDCARD)
        filename = act.utils.io_utils.pack_tar(met_files, write_directory=tmpdirname)
        filename = act.utils.io_utils.pack_gzip(filename, write_directory=tmpdirname, remove=True)
        ds = act.io.arm.read_arm_netcdf(filename)
        ds.clean.cleanup()

        assert 'temp_mean' in ds.data_vars

    with tempfile.TemporaryDirectory() as tmpdirname:
        met_files = sample_files.EXAMPLE_MET1
        filename = act.utils.io_utils.pack_gzip(met_files, write_directory=tmpdirname, remove=False)
        ds = act.io.arm.read_arm_netcdf(filename)
        ds.clean.cleanup()

        assert 'temp_mean' in ds.data_vars


def test_read_netcdf_tarfiles():
    with tempfile.TemporaryDirectory() as tmpdirname:
        met_files = list(Path(file) for file in act.tests.EXAMPLE_MET_WILDCARD)
        filename = act.utils.io_utils.pack_tar(met_files, write_directory=tmpdirname)
        ds = act.io.arm.read_arm_netcdf(filename)
        ds.clean.cleanup()

        assert 'temp_mean' in ds.data_vars


def test_unpack_tar():
    with tempfile.TemporaryDirectory() as tmpdirname:
        tar_file = Path(tmpdirname, 'tar_file_dir')
        output_dir = Path(tmpdirname, 'output_dir')
        tar_file.mkdir(parents=True, exist_ok=True)
        output_dir.mkdir(parents=True, exist_ok=True)

        for tar_file_name in ['test_file1.tar', 'test_file2.tar']:
            filenames = []
            for value in range(0, 10):
                filename = ''.join(random.choices(list(ascii_letters), k=15))
                filename = Path(tar_file, f'{filename}.nc')
                filename.touch()
                filenames.append(filename)
            act.utils.io_utils.pack_tar(
                filenames, write_filename=Path(tar_file, tar_file_name), remove=True
            )

        tar_files = list(tar_file.glob('*.tar'))
        result = act.utils.io_utils.unpack_tar(tar_files[0], write_directory=output_dir)
        assert isinstance(result, list)
        assert len(result) == 10
        for file in result:
            assert isinstance(file, (str, PathLike))

        files = list(output_dir.glob('*'))
        assert len(files) == 1
        assert files[0].is_dir()
        act.utils.io_utils.cleanup_files(dirname=output_dir)
        files = list(output_dir.glob('*'))
        assert len(files) == 0

        # Check not returing file but directory
        result = act.utils.io_utils.unpack_tar(
            tar_files[0], write_directory=output_dir, return_files=False
        )
        assert isinstance(result, str)
        files = list(Path(result).glob('*'))
        assert len(files) == 10
        act.utils.io_utils.cleanup_files(result)
        files = list(Path(output_dir).glob('*'))
        assert len(files) == 0

        # Test temporary directory
        result = act.utils.io_utils.unpack_tar(tar_files[0], temp_dir=True)
        assert isinstance(result, list)
        assert len(result) == 10
        for file in result:
            assert isinstance(file, (str, PathLike))

        act.utils.io_utils.cleanup_files(files=result)

        # Test removing TAR file
        result = act.utils.io_utils.unpack_tar(tar_files, write_directory=output_dir, remove=True)
        assert isinstance(result, list)
        assert len(result) == 20
        for file in result:
            assert isinstance(file, (str, PathLike))

        tar_files = list(tar_file.glob('*.tar'))
        assert len(tar_files) == 0

        act.utils.io_utils.cleanup_files(files=result)
        files = list(Path(output_dir).glob('*'))
        assert len(files) == 0

        not_a_tar_file = Path(tar_file, 'not_a_tar_file.tar')
        not_a_tar_file.touch()
        result = act.utils.io_utils.unpack_tar(not_a_tar_file, Path(output_dir, 'another_dir'))
        assert result == []

        act.utils.io_utils.cleanup_files()

        not_a_directory = '/asasfdlkjsdfjioasdflasdfhasd/not/a/directory'
        act.utils.io_utils.cleanup_files(dirname=not_a_directory)

        not_a_file = Path(not_a_directory, 'not_a_file.nc')
        act.utils.io_utils.cleanup_files(files=not_a_file)

        act.utils.io_utils.cleanup_files(files=output_dir)

        dir_names = list(Path(tmpdirname).glob('*'))
        for dir_name in [tar_file, output_dir]:
            assert dir_name, dir_name in dir_names

        filename = ''.join(random.choices(list(ascii_letters), k=15))
        filename = Path(tar_file, f'{filename}.nc')
        filename.touch()
        result = act.utils.io_utils.pack_tar(
            filename, write_filename=Path(tar_file, 'test_file_single'), remove=True
        )
        assert Path(filename).is_file() is False
        assert Path(result).is_file()
        assert result.endswith('.tar')


def test_gunzip():
    with tempfile.TemporaryDirectory() as tmpdirname:
        filenames = []
        for value in range(0, 10):
            filename = ''.join(random.choices(list(ascii_letters), k=15))
            filename = Path(tmpdirname, f'{filename}.nc')
            filename.touch()
            filenames.append(filename)

        filename = act.utils.io_utils.pack_tar(filenames, write_directory=tmpdirname, remove=True)
        files = list(Path(tmpdirname).glob('*'))
        assert len(files) == 1
        assert files[0].name == 'created_tarfile.tar'
        assert Path(filename).name == 'created_tarfile.tar'

        gzip_file = act.utils.io_utils.pack_gzip(filename=filename)
        files = list(Path(tmpdirname).glob('*'))
        assert len(files) == 2
        files = list(Path(tmpdirname).glob('*.gz'))
        assert files[0].name == 'created_tarfile.tar.gz'
        assert Path(gzip_file).name == 'created_tarfile.tar.gz'

        unpack_filename = act.utils.io_utils.unpack_gzip(filename=gzip_file)
        files = list(Path(tmpdirname).glob('*'))
        assert len(files) == 2
        assert Path(unpack_filename).name == 'created_tarfile.tar'

        result = act.utils.io_utils.unpack_tar(unpack_filename, return_files=True, randomize=True)
        files = list(Path(Path(result[0]).parent).glob('*'))
        assert len(result) == 10
        assert len(files) == 10
        for file in result:
            assert file.endswith('.nc')

    with tempfile.TemporaryDirectory() as tmpdirname:
        filenames = []
        for value in range(0, 10):
            filename = ''.join(random.choices(list(ascii_letters), k=15))
            filename = Path(tmpdirname, f'{filename}.nc')
            filename.touch()
            filenames.append(filename)

        filename = act.utils.io_utils.pack_tar(filenames, write_directory=tmpdirname, remove=True)
        files = list(Path(tmpdirname).glob('*'))
        assert len(files) == 1
        files = list(Path(tmpdirname).glob('*.tar'))
        assert files[0].name == 'created_tarfile.tar'
        assert Path(filename).name == 'created_tarfile.tar'

        gzip_file = act.utils.io_utils.pack_gzip(
            filename=filename, write_directory=Path(filename).parent, remove=False
        )
        files = list(Path(tmpdirname).glob('*'))
        assert len(files) == 2
        files = list(Path(tmpdirname).glob('*gz'))
        assert files[0].name == 'created_tarfile.tar.gz'
        assert Path(gzip_file).name == 'created_tarfile.tar.gz'

        unpack_filename = act.utils.io_utils.unpack_gzip(
            filename=gzip_file, write_directory=Path(filename).parent, remove=False
        )
        files = list(Path(tmpdirname).glob('*'))
        assert len(files) == 2
        assert Path(unpack_filename).name == 'created_tarfile.tar'

        result = act.utils.io_utils.unpack_tar(
            unpack_filename, return_files=True, randomize=False, remove=True
        )
        files = list(Path(Path(result[0]).parent).glob('*.nc'))
        assert len(result) == 10
        assert len(files) == 10
        for file in result:
            assert file.endswith('.nc')

        assert Path(unpack_filename).is_file() is False


@pytest.mark.skipif(not MOVIEPY_AVAILABLE, reason='MoviePy is not installed.')
def test_generate_movie():
    files = [
        'https://github.com/ARM-DOE/ACT/blob/main/tests/plotting/baseline/test_contour.png?raw=true',
        'https://github.com/ARM-DOE/ACT/blob/main/tests/plotting/baseline/test_contour2.png?raw=true',
        'https://github.com/ARM-DOE/ACT/blob/main/tests/plotting/baseline/test_contourf.png?raw=true',
        'https://github.com/ARM-DOE/ACT/blob/main/tests/plotting/baseline/test_contourf2.png?raw=true',
    ]
    cwd = Path.cwd()
    with tempfile.TemporaryDirectory() as tmpdirname:
        try:
            chdir(tmpdirname)

            # Test URL path for making movie
            result = act.utils.generate_movie(files, fps=5)
            assert Path(result).name == 'movie.mp4'

            # Test list of files for making movie
            files = [
                'test_contour.png',
                'test_contour2.png',
                'test_contourf.png',
                'test_contourf2.png',
            ]
            basepath = Path(Path(__file__).parents[1], 'plotting', 'baseline')
            files = [Path(basepath, fl) for fl in files]
            write_filename = Path(tmpdirname, 'one', 'two', 'three', 'movie_filename_testing.mp4')
            result = act.utils.generate_movie(files, write_filename=write_filename)
            assert result == str(write_filename)
            assert np.isclose(Path(write_filename).stat().st_size, 173189, 1000)

            # Test PosixPath generator for making movie
            result = act.utils.generate_movie(
                list(basepath.glob('test_contour[!_]*.png')), write_filename=write_filename
            )
            assert result == str(write_filename)
            assert np.isclose(Path(write_filename).stat().st_size, 173189, 1000)

            # Test passing path to directory of images
            dir_path = Path(tmpdirname, 'images')
            dir_path.mkdir()
            for fl in files:
                shutil.copy(fl, Path(dir_path, Path(fl).name))

            files = dir_path.glob('*.*')
            result = act.utils.generate_movie(dir_path)
            assert Path(result).name == 'movie.mp4'
            assert np.isclose(Path(result).stat().st_size, 173189, 1000)

            # Test converting movie format
            write_filename = 'movie2.mp4'
            result = act.utils.generate_movie(result, write_filename=write_filename)
            assert Path(result).name == write_filename
            assert np.isclose(Path(result).stat().st_size, 173189, 1000)

            # Test converting MPEG to mp4
            write_filename = 'movie3.mp4'
            mpeg_file = sample_files.EXAMPLE_MPEG
            result = act.utils.generate_movie(mpeg_file, write_filename=write_filename)
            files = list(Path().glob(write_filename))
            assert len(files) == 1
            assert np.isclose(files[0].stat().st_size, 1625298, rtol=100, atol=100)

        finally:
            chdir(cwd)


def test_arm_standards_validator():
    met_files = sample_files.EXAMPLE_MET_SAIL
    errors = act.utils.arm_standards_validator(met_files)
    assert len(errors) == 0

    ds = act.io.read_arm_netcdf(met_files)
    ds2 = ds.drop_vars(['lat', 'lon', 'alt'])
    errors = act.utils.arm_standards_validator(dataset=ds2)

    assert len(errors) == 3

    ds2 = ds
    var = ['lat', 'lon', 'alt']
    for v in var:
        del ds2[v].attrs['standard_name']

    errors = act.utils.arm_standards_validator(dataset=ds2)
    assert len(errors) == 3

    ds2 = ds
    for v in var:
        ds2[v].attrs['standard_name'] = 'test'
    errors = act.utils.arm_standards_validator(dataset=ds2)
    assert len(errors) == 3

    ds2 = ds
    for v in ds2:
        del ds2[v].attrs['long_name']
    errors = act.utils.arm_standards_validator(dataset=ds2)
    assert len(errors) == 54

    ds2 = act.io.read_arm_netcdf(met_files)
    ds2['time'].values[1] = ds2['time'].values[10]
    errors = act.utils.arm_standards_validator(dataset=ds2)

    assert 'Duplicate' in errors[0]
    assert 'increasing' in errors[1]

    file = 'shpinstrumentX50.z1.202005.000000.nc'
    errors = act.utils.arm_standards_validator(file)

    assert 'File is not in a standard format' in errors[0]

    file = 'sgpmetE13.z1.20200501.000000.nc'
    errors = act.utils.arm_standards_validator(file)

    assert 'no files to open' in errors[0]

    ds = act.io.read_arm_netcdf(sample_files.EXAMPLE_CEIL1)
    errors = act.utils.arm_standards_validator(dataset=ds)
    assert len(errors) == 4



"""
Functions for radiosonde related calculations.

"""

import numpy as np
import xarray as xr
from operator import itemgetter
from itertools import groupby
import metpy.calc as mpcalc
from metpy.units import units

from act.utils.data_utils import convert_to_potential_temp


def calculate_precipitable_water(ds, temp_name='tdry', rh_name='rh', pres_name='pres'):
    """

    Function to calculate precipitable water vapor from ARM sondewnpn b1 data.
    Will first calculate saturation vapor pressure of all data using Arden-Buck
    equations, then calculate specific humidity and integrate over all pressure
    levels to give us a precipitable water value in centimeters.

    ds : xarray.Dataset
        Xarray dataset as read in by the ACT netCDF reader.
    temp_name : str
        Name of temperature field to use. Defaults to 'tdry' for sondewnpn b1
        level data.
    rh_name : str
        Name of relative humidity field to use. Defaults to 'rh' for sondewnpn
        b1 level data.
    pres_name : str
        Name of atmospheric pressure field to use. Defaults to 'pres' for
        sondewnpn b1 level data.

    """
    temp = ds[temp_name].values
    rh = ds[rh_name].values
    pres = ds[pres_name].values

    # Get list of temperature values for saturation vapor pressure calc
    temperature = []
    for t in np.nditer(temp):
        temperature.append(t)

    # Apply Arden-Buck equation to get saturation vapor pressure
    sat_vap_pres = []
    for t in temperature:
        # Over liquid water, above freezing
        if t >= 0:
            sat_vap_pres.append(0.61121 * np.exp((18.678 - (t / 234.5)) * (t / (257.14 + t))))
        # Over ice, below freezing
        else:
            sat_vap_pres.append(0.61115 * np.exp((23.036 - (t / 333.7)) * (t / (279.82 + t))))

    # convert rh from % to decimal
    rel_hum = []
    for r in np.nditer(rh):
        rel_hum.append(r / 100.0)

    # get vapor pressure from rh and saturation vapor pressure
    vap_pres = []
    for i in range(0, len(sat_vap_pres)):
        es = rel_hum[i] * sat_vap_pres[i]
        vap_pres.append(es)

    # Get list of pressure values for mixing ratio calc
    pressure = []
    for p in np.nditer(pres):
        pressure.append(p)

    # Mixing ratio calc

    mix_rat = []
    for i in range(0, len(vap_pres)):
        mix_rat.append(0.622 * vap_pres[i] / (pressure[i] - vap_pres[i]))

    # Specific humidity

    spec_hum = []
    for rat in mix_rat:
        spec_hum.append(rat / (1 + rat))

    # Integrate specific humidity

    pwv = 0.0
    for i in range(1, len(pressure) - 1):
        pwv = pwv + 0.5 * (spec_hum[i] + spec_hum[i - 1]) * (pressure[i - 1] - pressure[i])

    pwv = pwv / 0.098
    return pwv


def calculate_stability_indicies(
    ds,
    temp_name='temperature',
    td_name='dewpoint_temperature',
    p_name='pressure',
    moving_ave_window=0,
):
    """
    Function for calculating stability indices from sounding data.

    Parameters
    ----------
    ds : ACT dataset
        The dataset to compute the stability indicies of. Must have
        temperature, dewpoint, and pressure in vertical coordinates.
    temp_name : str
        The name of the temperature field.
    td_name : str
        The name of the dewpoint field.
    p_name : str
        The name of the pressure field.
    moving_ave_window : int
        Number of points to do a moving average on sounding data to reduce
        noise. This is useful if noise in the sounding is preventing parcel
        ascent.

    Returns
    -------
    ds : ACT dataset
        An ACT dataset with additional stability indicies added.

    """
    t = ds[temp_name]
    td = ds[td_name]
    p = ds[p_name]

    if not hasattr(t, 'units'):
        raise AttributeError('Temperature field must have units' + ' for ACT to discern!')

    if not hasattr(td, 'units'):
        raise AttributeError('Dewpoint field must have units' + ' for ACT to discern!')

    if not hasattr(p, 'units'):
        raise AttributeError('Pressure field must have units' + ' for ACT to discern!')
    if t.units == 'C':
        t_units = units.degC
    else:
        t_units = getattr(units, t.units)

    if td.units == 'C':
        td_units = units.degC
    else:
        td_units = getattr(units, td.units)

    p_units = getattr(units, p.units)

    # Sort all values by decreasing pressure
    t_sorted = np.array(t.values)
    td_sorted = np.array(td.values)
    p_sorted = np.array(p.values)
    ind_sort = np.argsort(p_sorted)
    t_sorted = t_sorted[ind_sort[-1:0:-1]]
    td_sorted = td_sorted[ind_sort[-1:0:-1]]
    p_sorted = p_sorted[ind_sort[-1:0:-1]]

    if moving_ave_window > 0:
        t_sorted = np.convolve(t_sorted, np.ones((moving_ave_window,)) / moving_ave_window)
        td_sorted = np.convolve(td_sorted, np.ones((moving_ave_window,)) / moving_ave_window)
        p_sorted = np.convolve(p_sorted, np.ones((moving_ave_window,)) / moving_ave_window)

    t_sorted = t_sorted * t_units
    td_sorted = td_sorted * td_units
    p_sorted = p_sorted * p_units

    t_profile = mpcalc.parcel_profile(p_sorted, t_sorted[0], td_sorted[0])

    # Calculate parcel trajectory
    ds['parcel_temperature'] = t_profile.magnitude
    ds['parcel_temperature'].attrs['units'] = t_profile.units

    # Calculate CAPE, CIN, LCL
    sbcape, sbcin = mpcalc.surface_based_cape_cin(p_sorted, t_sorted, td_sorted)

    lcl = mpcalc.lcl(p_sorted[0], t_sorted[0], td_sorted[0])
    try:
        lfc = mpcalc.lfc(p_sorted[0], t_sorted[0], td_sorted[0])
    except IndexError:
        lfc = np.nan * p_sorted.units

    mucape, mucin = mpcalc.most_unstable_cape_cin(p_sorted, t_sorted, td_sorted)

    where_500 = np.argmin(np.abs(p_sorted - 500 * units.hPa))
    li = t_sorted[where_500] - t_profile[where_500]

    ds['surface_based_cape'] = sbcape.magnitude
    ds['surface_based_cape'].attrs['units'] = 'J/kg'
    ds['surface_based_cape'].attrs['long_name'] = 'Surface-based CAPE'
    ds['surface_based_cin'] = sbcin.magnitude
    ds['surface_based_cin'].attrs['units'] = 'J/kg'
    ds['surface_based_cin'].attrs['long_name'] = 'Surface-based CIN'
    ds['most_unstable_cape'] = mucape.magnitude
    ds['most_unstable_cape'].attrs['units'] = 'J/kg'
    ds['most_unstable_cape'].attrs['long_name'] = 'Most unstable CAPE'
    ds['most_unstable_cin'] = mucin.magnitude
    ds['most_unstable_cin'].attrs['units'] = 'J/kg'
    ds['most_unstable_cin'].attrs['long_name'] = 'Most unstable CIN'
    ds['lifted_index'] = li.magnitude
    ds['lifted_index'].attrs['units'] = t_profile.units
    ds['lifted_index'].attrs['long_name'] = 'Lifted index'
    ds['level_of_free_convection'] = lfc.magnitude
    ds['level_of_free_convection'].attrs['units'] = lfc.units
    ds['level_of_free_convection'].attrs['long_name'] = 'Level of free convection'
    ds['lifted_condensation_level_temperature'] = lcl[1].magnitude
    ds['lifted_condensation_level_temperature'].attrs['units'] = lcl[1].units
    ds['lifted_condensation_level_temperature'].attrs[
        'long_name'
    ] = 'Lifted condensation level temperature'
    ds['lifted_condensation_level_pressure'] = lcl[0].magnitude
    ds['lifted_condensation_level_pressure'].attrs['units'] = lcl[0].units
    ds['lifted_condensation_level_pressure'].attrs[
        'long_name'
    ] = 'Lifted condensation level pressure'
    return ds


def calculate_pbl_liu_liang(
    ds,
    temperature='tdry',
    pressure='pres',
    windspeed='wspd',
    height='alt',
    smooth_height=3,
    land_parameter=True,
    llj_max_alt=1500.0,
    llj_max_wspd=2.0,
):
    """
    Function for calculating the PBL height from a radiosonde profile
    using the Liu-Liang 2010 technique.  There are some slight descrepencies
    in the function from the ARM implementation 1.) it imposes a 1500m (keyword)
    height on the definition of the LLJ and 2.) the interpolation is slightly different
    using python functions

    Parameters
    ----------
    ds : xarray Dataset
        Dataset housing radiosonde profile for calculations
    temperature : str
        The name of the temperature field.
    pressure : str
        The name of the pressure field.
    windspeed : str
        The name of the  wind speed field.
    height : str
        The name of the height field
    smooth_height : int
        Number of points to do a moving average on sounding height data to reduce noise
    land_parameter : boolean
        Set to True if retrievals over land or false to retrievals over water
    llj_max_alt : float
        Maximum altitude the LLJ 2 m/s difference should be checked against
    llj_max_wspd : float
        Maximum wind speed threshold to use to define LLJ

    Returns
    -------
    ds : xarray Dataset
        xarray dataset with results stored in pblht_liu_liang variable

    References
    ----------
    Liu, Shuyan, and Xin-Zhong Liang. "Observed diurnal cycle climatology of planetary
        boundary layer height." Journal of Climate 23, no. 21 (2010): 5790-5809.

    Sivaraman, C., S. McFarlane, E. Chapman, M. Jensen, T. Toto, S. Liu, and M. Fischer.
        "Planetary boundary layer (PBL) height value added product (VAP): Radiosonde retrievals."
        Department of Energy Office of Science Atmospheric Radiation Measurement (ARM) Program
        (United States) (2013).

    """

    # Preprocess the sonde data to ensure the same methods across all retrievals
    ds2 = preprocess_sonde_data(
        ds,
        temperature=temperature,
        pressure=pressure,
        height=height,
        smooth_height=smooth_height,
        base=5.0,
    )

    pres = ds2[pressure].values
    wspd = ds2[windspeed].values
    alt = ds2[height].values

    theta = ds2['potential_temperature'].values

    # Calculate the lapse rate
    theta_gradient = np.diff(theta) / np.diff(alt)

    # Calculate AGL
    if np.isnan(alt[0]):
        idx = np.where(~np.isnan(alt))[0]
        agl = alt - alt[idx[0]]
    else:
        agl = alt - alt[0]

    theta_diff = theta[4] - theta[1]
    theta_gradient = np.diff(theta) / np.diff(alt / 1000.0)

    # Set up threshold values
    if land_parameter:
        stability_thresh = 1.0  # K
        inst_thresh = 0.5  # K
        overshoot_thresh = 4.0  # K/km
    else:
        stability_thresh = 0.2  # K
        inst_thresh = 0.1  # K
        overshoot_thresh = 0.5  # K/km

    # Check Regimes
    if theta_diff < 0 - stability_thresh:
        regime = 'CBL'
    if theta_diff > abs(stability_thresh):
        regime = 'SBL'
    if (0 - stability_thresh) <= theta_diff <= abs(stability_thresh):
        regime = 'NRL'

    # Calculate for CBL/NRL regimes
    pbl_stable = np.nan
    pbl_shear = np.nan

    if regime == 'CBL' or regime == 'NRL':
        # Calculate gradient from first level
        theta_gradient_0 = theta - theta[0]

        # Only process data above 150m ARM
        idx = np.where(agl > 150)[0][0]
        theta_gradient_0[0:idx] = np.nan

        # Scan upward to find lowest level that meets condition
        idx = np.where(theta_gradient_0 >= inst_thresh)[0]
        theta_gradient[0 : idx[0]] = np.nan

        # Scan upward from previous level to search for overlying inversion layer
        idx = np.where(theta_gradient >= overshoot_thresh)[0]
        pbl = alt[idx[0]]
    else:
        idx = np.array(
            [
                i
                for i, t in enumerate(theta_gradient[1:-1])
                if theta_gradient[i] < theta_gradient[i - 1]
                and theta_gradient[i] < theta_gradient[i + 1]
            ]
        )

        for i in idx:
            cond1 = (theta_gradient[i] - theta_gradient[i - 1]) < -40.0
            cond2 = (theta_gradient[i + 1] < overshoot_thresh) or (
                theta_gradient[i + 2] < overshoot_thresh
            )
            if cond1 or cond2:
                # This gets the ARM answer
                pbl_stable = (alt[i + 1] + alt[i]) / 2.0
                # pbl_stable = alt[i]
                break

        # Check for low-level jet
        # Find the height of the maximum windspeed and look up to find layer 2m/s lower
        # Stull 1988 indicates LLJ is defined as where there is a relative wind speed
        # maximum that is more than 2 m/s faster than the wind speeds above it within
        # the lowest 1500m of the atmosphere. Keywords to adjust are provided
        idh = np.where(alt <= llj_max_alt)[0]
        max_wspd_ind = [i for i, w in enumerate(wspd[:-1]) if wspd[i] > wspd[i + 1]][0]
        diff = wspd[max_wspd_ind] - wspd[max_wspd_ind : idh[-1]]
        idx = np.where(diff > llj_max_wspd)[0]
        if len(idx) > 0:
            wspd_to_surf = np.diff(np.flip(wspd[0:max_wspd_ind]))
            wspd_monotonic = np.all(wspd_to_surf <= 0.0)
            if wspd_monotonic:
                pbl_shear = alt[max_wspd_ind]

        if ~np.all(np.isnan([pbl_stable, pbl_shear])):
            pbl = np.nanmin([pbl_stable, pbl_shear])
        else:
            pbl = -9999.0

    atts = {'units': 'm', 'long_name': 'Planteary Boundary Layer Height Liu-Liang'}
    da = xr.DataArray(pbl, attrs=atts)
    ds['pblht_liu_liang'] = da

    atts = {
        'units': '',
        'long_name': 'Planteary Boundary Layer Regime Classification Liu-Liang',
    }
    da = xr.DataArray(regime, attrs=atts)
    ds['pblht_regime_liu_liang'] = da

    atts = {'units': 'mb', 'long_name': 'Gridded pressure'}
    da = xr.DataArray(pres, coords={'atm_pres_ss': pres}, dims=['atm_pres_ss'], attrs=atts)
    ds['atm_pres_ss'] = da

    atts = {'units': 'K', 'long_name': 'Gridded potential temperature'}
    da = xr.DataArray(theta, coords={'atm_pres_ss': pres}, dims=['atm_pres_ss'], attrs=atts)
    ds['potential_temperature_ss'] = da

    atts = {'units': 'm', 'long_name': 'Gridded altitude'}
    da = xr.DataArray(alt, coords={'atm_pres_ss': pres}, dims=['atm_pres_ss'], attrs=atts)
    ds['alt_ss'] = da

    atts = {'units': 'm', 'long_name': 'PBL Stable Condition 1'}
    da = xr.DataArray(pbl_stable, attrs=atts)
    ds['pblht_liu_liang_stable_cond'] = da

    atts = {'units': 'm', 'long_name': 'PBL Shear Condition 2'}
    da = xr.DataArray(pbl_shear, attrs=atts)
    ds['pblht_liu_liang_shear_cond'] = da

    return ds


def calculate_pbl_heffter(
    ds,
    temperature='tdry',
    pressure='pres',
    height='alt',
    smooth_height=3,
    base=5.0,
):
    """
    Function for calculating the PBL height from a radiosonde profile
    using the Heffter technique. There are differences from the ARM
    VAP at times due to different averaging schemes.  Larger differences
    do occur at times and are unknown as to the cause but it is being
    investigated and is potential a code issue with the VAP.

    Parameters
    ----------
    ds : xarray Dataset
        Dataset housing radiosonde profile for calculations
    temperature : str
        The name of the temperature field.
    pressure : str
        The name of the pressure field.
    height : str
        The name of the height field
    smooth_height : int
        Number of points to do a moving average on sounding height data to reduce noise
    base : int
        Interval for pressure gridding.  In testing, 5 mb was found to produce results with
        the lowest RMS

    Returns
    -------
    ds : xarray Dataset
        xarray dataset with results stored in pblht_liu_liang variable

    References
    ----------
    Heffter JL. 1980. “Transport Layer Depth Calculations.” Second Joint Conference on
        Applications of Air Pollution Meteorology, New Orleans, Louisiana.

    Sivaraman, C., S. McFarlane, E. Chapman, M. Jensen, T. Toto, S. Liu, and M. Fischer.
        "Planetary boundary layer (PBL) height value added product (VAP): Radiosonde retrievals."
        Department of Energy Office of Science Atmospheric Radiation Measurement (ARM) Program
        (United States) (2013).

    """

    # Preprocess the sonde data to ensure the same methods across all retrievals
    ds2 = preprocess_sonde_data(
        ds,
        temperature=temperature,
        pressure=pressure,
        height=height,
        smooth_height=smooth_height,
        base=base,
    )

    # Get data
    pres = ds2[pressure].values
    alt = ds2[height].values
    theta = ds2['potential_temperature'].values

    # Calculate the lapse rate
    theta_gradient = np.diff(theta) / np.diff(alt)

    # Calculate AGL
    if np.isnan(alt[0]):
        idx = np.where(~np.isnan(alt))[0]
        agl = alt - alt[idx[0]]
    else:
        agl = alt - alt[0]

    # Find where the lapse rate is greater than 0.005 K/m
    idx = np.where(theta_gradient >= 0.005)[0]

    # Find the consistent layers by grouping the indices together
    # Does not include a single height as a layer
    ranges = []
    for key, group in groupby(enumerate(idx), lambda i: i[0] - i[1]):
        group = list(map(itemgetter(1), group))
        if group[-1] - group[0] > 0:
            ranges.append((group[0], group[-1]))

    # Subset ranges to lowest 5
    if len(ranges) > 5:
        ranges = ranges[0:5]

    # For each layer, calculate the difference in theta from
    # top and bottom of the layer.  The lowest layer where the
    # difference is > 2 K is set as the PBL.
    pbl = 0.0
    theta_diff_layer = []
    bottom_inversion = []
    top_inversion = []
    for r in ranges:
        if agl[r[1]] > 4000.0:
            continue
        theta_diff = theta[r[1]] - theta[r[0]]
        theta_diff_layer.append(theta_diff)
        bottom_inversion.append(alt[r[0]])
        top_inversion.append(alt[r[1]])
        if pbl == 0.0 and theta_diff > 2.0:
            pbl = alt[r[0]]

    if len(theta_diff_layer) == 0:
        pbl = -9999.0

    # If PBL is not set, set it to the layer with the max theta diff
    if pbl == 0.0:
        idx = np.argmax(theta_diff_layer)
        pbl = bottom_inversion[idx]

    # Add variables to the dataset
    atts = {'units': 'm', 'long_name': 'Planteary Boundary Layer Height Heffter'}
    da = xr.DataArray(pbl, attrs=atts)
    ds['pblht_heffter'] = da

    atts = {'units': 'mb', 'long_name': 'Gridded pressure'}
    da = xr.DataArray(pres, coords={'atm_pres_ss': pres}, dims=['atm_pres_ss'], attrs=atts)
    ds['atm_pres_ss'] = da

    atts = {'units': 'K', 'long_name': 'Gridded potential temperature'}
    da = xr.DataArray(theta, coords={'atm_pres_ss': pres}, dims=['atm_pres_ss'], attrs=atts)
    ds['potential_temperature_ss'] = da

    atts = {'units': 'm', 'long_name': 'Gridded altitude'}
    da = xr.DataArray(alt, coords={'atm_pres_ss': pres}, dims=['atm_pres_ss'], attrs=atts)
    ds['alt_ss'] = da

    atts = {'units': 'm', 'long_name': 'Bottom height of inversion layers'}
    da = xr.DataArray(
        bottom_inversion,
        coords={'layers': list(range(len(bottom_inversion)))},
        dims=['layers'],
        attrs=atts,
    )
    ds['bottom_inversion'] = da

    atts = {'units': 'm', 'long_name': 'Top height of inversion layers'}
    da = xr.DataArray(
        top_inversion,
        coords={'layers': list(range(len(top_inversion)))},
        dims=['layers'],
        attrs=atts,
    )
    ds['top_inversion'] = da

    return ds


def preprocess_sonde_data(
    ds,
    temperature='tdry',
    pressure='pres',
    height='alt',
    smooth_height=3,
    base=5.0,
):
    """
    Function for processing the SONDE data for the PBL calculations.
    This is to ensure consistency and also applies some QC to the
    processing.

    Parameters
    ----------
    ds : xarray Dataset
        Dataset housing radiosonde profile for calculations
    temperature : str
        The name of the temperature field.
    pressure : str
        The name of the pressure field.
    height : str
        The name of the height field
    smooth_height : int
        Number of points to do a moving average on sounding height data to reduce noise
    base : int
        Interval for pressure gridding.  In testing, 5 mb was found to produce results with
        the lowest RMS

    Returns
    -------
    ds : xarray.Dataset
        Xarray dataset containing processed sonde data.

    References
    ----------
    Sivaraman, C., S. McFarlane, E. Chapman, M. Jensen, T. Toto, S. Liu, and M. Fischer.
        "Planetary boundary layer (PBL) height value added product (VAP): Radiosonde retrievals."
        Department of Energy Office of Science Atmospheric Radiation Measurement (ARM) Program
        (United States) (2013).

    """

    # Get the initial time and temp values
    time_0 = ds['time'].values
    temp_0 = ds[temperature].values

    # Apply a rolling average to smooth the pressure out
    ds[pressure] = ds[pressure].rolling(time=smooth_height, min_periods=1, center=True).mean()

    # Swap time and pressure for doing the appropriate gridding
    ds2 = ds.swap_dims(dims_dict={'time': pressure})
    for var in ds2:
        ds2[var].attrs = ds2[var].attrs

    # Set up the pressure grid
    starting_pres = base * np.ceil(float(ds2[pressure].values[2]) / base)
    p_grid = np.flip(np.arange(100.0, starting_pres + base, base))

    # Pull out the data that's nearest to the pressure grid.  If it errors
    # it will smooth the data more.  This tends to happen if there are multiple
    # values of pressure that are the same
    try:
        ds2 = ds2.sel({pressure: p_grid}, method='nearest')
    except Exception:
        ds[pressure] = (
            ds[pressure].rolling(time=smooth_height + 4, min_periods=2, center=True).mean()
        )
        ds2 = ds.swap_dims(dims_dict={'time': pressure})
        for var in ds2:
            ds2[var].attrs = ds[var].attrs
        try:
            ds2 = ds2.sel({pressure: p_grid}, method='nearest')
        except Exception:
            raise ValueError('Sonde profile does not have unique pressures after smoothing')

    # Get data
    alt = ds2[height].values
    pres = ds2[pressure].values
    temp = ds2[temperature].values

    # Perform Pre-processing checks
    if len(temp) == 0.0:
        raise ValueError('No data in profile')

    if np.nanmax(alt) < 1000.0:
        raise ValueError('Max altitude < 1000m')

    if np.nanmax(pres) <= 200.0:
        raise ValueError('Max pressure <= 200 hPa')

    # Check temperature delta
    t1 = time_0[0]
    t2 = t1 + np.timedelta64(10, 's')
    idx = np.where((time_0 >= t1) & (time_0 <= t2))[0]
    t_delta = abs(temp_0[idx[-1]] - temp_0[idx[0]])
    if t_delta > 30.0:
        raise ValueError('Temperature changes by >30º in first 10 seconds')

    # Check min/max
    if np.nanmax(temp) > 50.0 or np.nanmin(temp) < -90:
        raise ValueError('Temperature outside acceptable range (-90, 50)')

    if np.isnan(pres[0]) or np.isnan(pres[1]):
        raise ValueError('First two pressure values bad')

    # Calculate potential temperature and subsequent gradients
    theta = (
        convert_to_potential_temp(ds=ds2, temp_var_name=temperature, press_var_name=pressure)
        + 273.15
    )

    # Set variables to return
    atts = {'units': 'K', 'long_name': 'Potential temperature'}
    da = xr.DataArray(theta, coords=ds2['tdry'].coords, dims=ds2[temperature].dims, attrs=atts)
    ds2['potential_temperature'] = da

    return ds2
